import logging

from csrank.core.fate_linear import FATELinearCore
from csrank.losses import hinged_rank_loss
from .object_ranker import ObjectRanker


class FATELinearObjectRanker(FATELinearCore, ObjectRanker):
    def __init__(
        self,
        n_hidden_set_units=2,
        loss_function=hinged_rank_loss,
        learning_rate=1e-3,
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create a FATELinear-network architecture for leaning discrete choice function. The first-aggregate-then-evaluate
            approach learns an embedding of each object and then aggregates that into a context representation
            :math:`\\mu_{C(x)}` and then scores each object :math:`x` using a generalized utility function
            :math:`U (x, \\mu_{C(x)})`.
            To make it computationally efficient we take the the context :math:`C(x)` as query set :math:`Q`.
            The context-representation is evaluated as:

            .. math::
                \\mu_{C(x)} = \\frac{1}{\\lvert C(x) \\lvert} \\sum_{y \\in C(x)} \\phi(y)

            where :math:`\\phi \\colon \\mathcal{X} \\to \\mathcal{Z}` maps each object :math:`y` to an
            :math:`m`-dimensional embedding space :math:`\\mathcal{Z} \\subseteq \\mathbb{R}^m`.
            Training complexity is quadratic in the number of objects and prediction complexity is only linear.
            The discrete choice for the given query set :math:`Q` is defined as:

            .. math::

                dc(Q) := \\operatorname{argmax}_{x \\in Q}  \\;  U (x, \\mu_{C(x)})

            Parameters
            ----------
            n_hidden_set_units : int
                Number of hidden set units.
            batch_size : int
                Batch size to use for training
            loss_function : function
                Differentiable loss function for the score vector
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the @FATENetwork
        """
        super().__init__(
            n_hidden_set_units=n_hidden_set_units,
            learning_rate=learning_rate,
            batch_size=batch_size,
            loss_function=loss_function,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(FATELinearObjectRanker.__name__)

    def fit(
        self, X, Y, epochs=10, callbacks=None, validation_split=0.1, verbose=0, **kwd
    ):
        """
                    Fit an object ranking learning model on a provided set of queries.
                    The provided queries are of a fixed size (numpy arrays).

                    Parameters
                    ----------
                    X : numpy array
                        (n_instances, n_objects, n_features)
                        Feature vectors of the objects
                    Y : numpy array
                        (n_instances, n_objects)
                        Rankings of the given objects
                    epochs : int
                        Number of epochs to run if training for a fixed query size
                    callbacks : list
                        List of callbacks to be called during optimization
                    validation_split : float
                        Percentage of instances to split off to validate on
                    verbose : bool
                        Print verbose information
                    **kwd
                        Keyword arguments for the fit function
                """
        super().fit(
            X,
            Y,
            epochs=epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            verbose=verbose,
            **kwd,
        )

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ObjectRanker.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def set_tunable_parameters(
        self,
        n_hidden_set_units=32,
        learning_rate=1e-3,
        batch_size=128,
        epochs_drop=300,
        drop=0.1,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden_set_units=n_hidden_set_units,
            learning_rate=learning_rate,
            batch_size=batch_size,
            epochs_drop=epochs_drop,
            drop=drop,
            **point,
        )



from itertools import combinations
from itertools import permutations
import logging

from keras import backend as K
from keras import Input
from keras import Model
from keras.layers import Activation
from keras.layers import concatenate
from keras.layers import Dense
from keras.layers import Lambda
from keras.optimizers import SGD
from keras.regularizers import l2
import numpy as np

from csrank.core.feta_network import FETANetwork
from csrank.layers import NormalizedDense
from csrank.numpy_util import sigmoid
from .discrete_choice import DiscreteObjectChooser


class FETADiscreteChoiceFunction(FETANetwork, DiscreteObjectChooser):
    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        add_zeroth_order_model=False,
        max_number_of_objects=10,
        num_subsample=5,
        loss_function="categorical_hinge",
        batch_normalization=False,
        kernel_regularizer=l2(1e-4),
        kernel_initializer="lecun_normal",
        activation="selu",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=["categorical_accuracy"],
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create a FETA-network architecture for learning the discrete choice functions.
            The first-evaluate-then-aggregate approach approximates the context-dependent utility function using the
            first-order utility function :math:`U_1 \\colon \\mathcal{X} \\times \\mathcal{X} \\rightarrow [0,1]`
            and zeroth-order utility function  :math:`U_0 \\colon \\mathcal{X} \\rightarrow [0,1]`.
            The scores each object :math:`x` using a context-dependent utility function :math:`U (x, C_i)`:

            .. math::
                 U(x_i, C_i) = U_0(x_i) + \\frac{1}{n-1} \\sum_{x_j \\in Q \\setminus \\{x_i\\}} U_1(x_i , x_j) \\, .

            Training and prediction complexity is quadratic in the number of objects.
            The discrete choice for the given query set :math:`Q` is defined as:

            .. math::

                dc(Q) := \\operatorname{argmax}_{x_i \\in Q}  \\;  U (x_i, C_i)

            Parameters
            ----------
            n_hidden : int
                Number of hidden layers
            n_units : int
                Number of hidden units in each layer
            add_zeroth_order_model : bool
                True if the model should include a latent utility function
            max_number_of_objects : int
                The maximum number of objects to train from
            num_subsample : int
                Number of objects to subsample to
            loss_function : function
                Differentiable loss function for the score vector
            batch_normalization : bool
                Whether to use batch normalization in the hidden layers
            kernel_regularizer : function
                Regularizer to use in the hidden units
            kernel_initializer : function or string
                Initialization function for the weights of each hidden layer
            activation : string or function
                Activation function to use in the hidden units
            optimizer : string or function
                Stochastic gradient optimizer
            metrics : list
                List of evaluation metrics (can be non-differentiable)
            batch_size : int
                Batch size to use for training
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the hidden units
        """
        super().__init__(
            n_hidden=n_hidden,
            n_units=n_units,
            add_zeroth_order_model=add_zeroth_order_model,
            max_number_of_objects=max_number_of_objects,
            num_subsample=num_subsample,
            loss_function=loss_function,
            batch_normalization=batch_normalization,
            kernel_regularizer=kernel_regularizer,
            kernel_initializer=kernel_initializer,
            activation=activation,
            optimizer=optimizer,
            metrics=metrics,
            batch_size=batch_size,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(FETADiscreteChoiceFunction.__name__)

    def _construct_layers(self, **kwargs):
        self.input_layer = Input(
            shape=(self.n_objects_fit_, self.n_object_features_fit_)
        )
        # Todo: Variable sized input
        # X = Input(shape=(None, n_features))
        if self.batch_normalization:
            if self._use_zeroth_model:
                self.hidden_layers_zeroth = [
                    NormalizedDense(
                        self.n_units, name="hidden_zeroth_{}".format(x), *kwargs
                    )
                    for x in range(self.n_hidden)
                ]
            self.hidden_layers = [
                NormalizedDense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        else:
            if self._use_zeroth_model:
                self.hidden_layers_zeroth = [
                    Dense(self.n_units, name="hidden_zeroth_{}".format(x), **kwargs)
                    for x in range(self.n_hidden)
                ]
            self.hidden_layers = [
                Dense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        assert len(self.hidden_layers) == self.n_hidden
        self.output_node = Dense(
            1,
            activation="linear",
            kernel_regularizer=self.kernel_regularizer,
            name="score",
        )
        if self._use_zeroth_model:
            self.output_node_zeroth = Dense(
                1,
                activation="linear",
                kernel_regularizer=self.kernel_regularizer,
                name="zero_score",
            )
            self.weighted_sum = Dense(
                1,
                activation="sigmoid",
                kernel_regularizer=self.kernel_regularizer,
                name="weighted_sum",
            )

    def construct_model(self):
        """
            Construct the :math:`1`-st order and :math:`0`-th order models, which are used to approximate the
            :math:`U_1(x, C(x))` and the :math:`U_0(x)` utilities respectively. For each pair of objects in
            :math:`x_i, x_j \\in Q` :math:`U_1(x, C(x))` we construct :class:`CmpNetCore` with weight sharing to
            approximate a pairwise-matrix. A pairwise matrix with index (i,j) corresponds to the :math:`U_1(x_i,x_j)`
            is a measure of how favorable it is to choose :math:`x_i` over :math:`x_j`. Using this matrix we calculate
            the borda score for each object to calculate :math:`U_1(x, C(x))`. For `0`-th order model we construct
            :math:`\\lvert Q \\lvert` sequential networks whose weights are shared to evaluate the :math:`U_0(x)` for
            each object in the query set :math:`Q`. The output mode is using sigmoid activation.

            Returns
            -------
            model: keras :class:`Model`
                Neural network to learn the FETA utility score
        """

        def create_input_lambda(i):
            return Lambda(lambda x: x[:, i])

        if self._use_zeroth_model:
            self.logger.debug("Create 0th order model")
            zeroth_order_outputs = []
            inputs = []
            for i in range(self.n_objects_fit_):
                x = create_input_lambda(i)(self.input_layer)
                inputs.append(x)
                for hidden in self.hidden_layers_zeroth:
                    x = hidden(x)
                zeroth_order_outputs.append(self.output_node_zeroth(x))
            zeroth_order_scores = concatenate(zeroth_order_outputs)
            self.logger.debug("0th order model finished")
        self.logger.debug("Create 1st order model")
        outputs = [list() for _ in range(self.n_objects_fit_)]
        for i, j in combinations(range(self.n_objects_fit_), 2):
            if self._use_zeroth_model:
                x1 = inputs[i]
                x2 = inputs[j]
            else:
                x1 = create_input_lambda(i)(self.input_layer)
                x2 = create_input_lambda(j)(self.input_layer)
            x1x2 = concatenate([x1, x2])
            x2x1 = concatenate([x2, x1])

            for hidden in self.hidden_layers:
                x1x2 = hidden(x1x2)
                x2x1 = hidden(x2x1)

            merged_left = concatenate([x1x2, x2x1])
            merged_right = concatenate([x2x1, x1x2])

            N_g = self.output_node(merged_left)
            N_l = self.output_node(merged_right)

            outputs[i].append(N_g)
            outputs[j].append(N_l)
        # convert rows of pairwise matrix to keras layers:
        outputs = [concatenate(x) for x in outputs]

        # compute utility scores:
        scores = [
            Lambda(lambda s: K.mean(s, axis=1, keepdims=True))(x) for x in outputs
        ]
        scores = concatenate(scores)
        self.logger.debug("1st order model finished")
        if self._use_zeroth_model:

            def get_score_object(i):
                return Lambda(lambda x: x[:, i, None])

            concat_scores = [
                concatenate(
                    [
                        get_score_object(i)(scores),
                        get_score_object(i)(zeroth_order_scores),
                    ]
                )
                for i in range(self.n_objects_fit_)
            ]
            scores = []
            for i in range(self.n_objects_fit_):
                scores.append(self.weighted_sum(concat_scores[i]))
            scores = concatenate(scores)

        # if self._use_zeroth_model:
        #     scores = add([scores, zeroth_order_scores])
        # if self._use_zeroth_model:
        #     def expand_dims():
        #         return Lambda(lambda x: x[..., None])
        #
        #     def squeeze_dims():
        #         return Lambda(lambda x: x[:, :, 0])
        #
        #     scores = expand_dims()(scores)
        #     zeroth_order_scores = expand_dims()(zeroth_order_scores)
        #     concat_scores = concatenate([scores, zeroth_order_scores], axis=-1)
        #     weighted_sum = Conv1D(name='weighted_sum', filters=1, kernel_size=(1), strides=1, activation='linear',
        #                          kernel_initializer=self.kernel_initializer, input_shape=(self.n_objects_fit_, 2),
        #                          kernel_regularizer=self.kernel_regularizer, use_bias=False)
        #     scores = weighted_sum(concat_scores)
        #     scores = squeeze_dims()(scores)
        if not self._use_zeroth_model:
            scores = Activation("sigmoid")(scores)
        model = Model(inputs=self.input_layer, outputs=scores)
        self.logger.debug("Compiling complete model...")
        model.compile(
            loss=self.loss_function, optimizer=self.optimizer, metrics=self.metrics
        )
        return model

    def _create_weighted_model(self, n_objects):
        def get_score_object(i):
            return Lambda(lambda x: x[:, i, None])

        s1 = Input(shape=(n_objects,))
        s2 = Input(shape=(n_objects,))
        concat_scores = [
            concatenate([get_score_object(i)(s1), get_score_object(i)(s2)])
            for i in range(n_objects)
        ]
        scores = []
        for i in range(n_objects):
            scores.append(self.weighted_sum(concat_scores[i]))
        scores = concatenate(scores)
        model = Model(inputs=[s1, s2], outputs=scores)
        return model

    def _predict_scores_using_pairs(self, X, **kwd):
        n_instances, n_objects, n_features = X.shape
        n2 = n_objects * (n_objects - 1)
        pairs = np.empty((n2, 2, n_features))
        scores = np.zeros((n_instances, n_objects))
        for n in range(n_instances):
            for k, (i, j) in enumerate(permutations(range(n_objects), 2)):
                pairs[k] = (X[n, i], X[n, j])
            result = self._predict_pair(
                pairs[:, 0], pairs[:, 1], only_pairwise=True, **kwd
            )[:, 0]
            scores[n] += result.reshape(n_objects, n_objects - 1).mean(axis=1)
            del result
        del pairs
        if self._use_zeroth_model:
            scores_zero = self.zero_order_model.predict(X.reshape(-1, n_features))
            scores_zero = scores_zero.reshape(n_instances, n_objects)
            model = self._create_weighted_model(n_objects)
            scores = model.predict([scores, scores_zero], **kwd)
        else:
            scores = sigmoid(scores)
        return scores

    def _create_zeroth_order_model(self):
        inp = Input(shape=(self.n_object_features_fit_,))

        x = inp
        for hidden in self.hidden_layers_zeroth:
            x = hidden(x)
        zeroth_output = self.output_node_zeroth(x)

        return Model(inputs=[inp], outputs=Activation("sigmoid")(zeroth_output))

    def fit(self, X, Y, **kwd):
        super().fit(X, Y, **kwd)

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return DiscreteObjectChooser.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def clear_memory(self, **kwargs):
        self.logger.info("Clearing memory")
        super().clear_memory(**kwargs)

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden=n_hidden,
            n_units=n_units,
            reg_strength=reg_strength,
            learning_rate=learning_rate,
            batch_size=batch_size,
            **point,
        )



import numpy as np

from csrank.constants import DISCRETE_CHOICE
from .util import angle_between
from .util import convert_to_label_encoding
from ..mnist_dataset_reader import MNISTDatasetReader


class MNISTDiscreteChoiceDatasetReader(MNISTDatasetReader):
    def __init__(self, dataset_type="unique", **kwargs):
        dataset_func_dict = {
            "unique": self.create_dataset_unique,
            "largest": self.create_dataset_largest,
            "median": self.create_dataset_median,
            "unique_max_occurring": self.create_dataset_mode_least_angle,
        }
        if dataset_type not in dataset_func_dict:
            raise ValueError(
                f"dataset_type must be one of {set(dataset_func_dict.keys())}"
            )
        self.dataset_function = dataset_func_dict[dataset_type]
        super(MNISTDiscreteChoiceDatasetReader, self).__init__(
            learning_problem=DISCRETE_CHOICE, **kwargs
        )
        self.logger.info("Dataset type {}".format(dataset_type))

    def create_dataset_median(self):
        self.logger.info("create_dataset_median")
        if self.n_objects % 2 == 0:
            self.n_objects = self.n_objects - 1
            self.logger.info(
                "Cannot create the dataset for even numbered sets so decreasing the set size {}".format(
                    self.n_objects
                )
            )

        n_total = self.n_test_instances + self.n_train_instances
        self.X = np.empty((n_total, self.n_objects, self.n_features))
        self.Y = np.empty(n_total, dtype=int)
        all_indices = np.arange(len(self.X_raw))
        for i in range(n_total):
            choice = np.arange(3)
            while len(choice) != 1:
                indices = self.random_state.choice(
                    all_indices, size=self.n_objects, replace=False
                )
                labels = self.y_labels[indices]
                choice = np.where(labels == np.median(labels))[0]
                if len(choice) == 1:
                    self.X[i] = self.X_raw[indices]
                    self.Y[i] = choice
        self.Y = convert_to_label_encoding(self.Y, self.n_objects)
        self.__check_dataset_validity__()

    def create_dataset_largest(self):
        self.logger.info("create_dataset_largest new")
        n_total = self.n_test_instances + self.n_train_instances
        self.X = np.empty((n_total, self.n_objects, self.n_features))
        self.Y = np.empty(n_total, dtype=int)
        all_indices = np.arange(len(self.X_raw))
        for i in range(n_total):
            choice = np.arange(3)
            while len(choice) != 1:
                indices = self.random_state.choice(
                    all_indices, size=self.n_objects, replace=False
                )
                labels = self.y_labels[indices]
                choice = np.where(labels == np.max(labels))[0]
                if len(choice) == 1:
                    self.X[i] = self.X_raw[indices]
                    self.Y[i] = choice
        self.Y = convert_to_label_encoding(self.Y, self.n_objects)
        self.__check_dataset_validity__()

    def create_dataset_mode_least_angle(self):
        self.logger.info("create_dataset_maximum_occurring_number")
        n_total = self.n_test_instances + self.n_train_instances
        self.X = np.empty((n_total, self.n_objects, self.n_features))
        self.Y = np.empty(n_total, dtype=int)
        all_indices = np.arange(len(self.X_raw))
        weights = self.random_state.randn(self.n_features)
        for i in range(n_total):
            while True:
                indices = self.random_state.choice(
                    all_indices, size=self.n_objects, replace=False
                )
                labels = self.y_labels[indices]
                numbers, counts = np.unique(labels, return_counts=True)
                max_count = np.where(counts == np.max(counts))[0]
                if len(max_count) == 1:
                    self.X[i] = self.X_raw[indices]
                    largest_set = np.where(labels == numbers[max_count])[0]
                    scores = np.array(
                        [angle_between(x, weights) for x in self.X[i][largest_set]]
                    )
                    self.Y[i] = largest_set[np.argmax(scores)]
                    break
        self.Y = convert_to_label_encoding(self.Y, self.n_objects)
        self.__check_dataset_validity__()

    def create_dataset_unique(self):
        self.logger.info("create_dataset_unique")
        num_classes = len(np.unique(self.y_labels))
        n_total = self.n_test_instances + self.n_train_instances
        unique_number = self.random_state.randint(0, num_classes, size=n_total)
        self.X = np.empty((n_total, self.n_objects, self.n_features))
        self.Y = np.empty(n_total, dtype=int)
        classes = np.arange(num_classes)
        for i in range(n_total):
            s = int((self.n_objects - 1) / 2)
            choices = self.random_state.choice(
                np.delete(classes, unique_number[i]), size=s
            )
            ind = self.random_state.choice(
                np.where(self.y_labels == unique_number[i])[0], size=1
            )[0]
            indices = [ind]
            for j, c in enumerate(choices):
                remaining = np.where(self.y_labels == c)[0]
                size = 2
                if self.n_objects % 2 == 0 and j == len(choices) - 1:
                    size = 3
                ind = list(
                    self.random_state.choice(remaining, size=size, replace=False)
                )
                indices = indices + ind
            self.random_state.shuffle(indices)
            self.X[i] = self.X_raw[indices]
            position = np.where(self.y_labels[indices] == unique_number[i])
            self.Y[i] = position[0][0]
        self.Y = convert_to_label_encoding(self.Y, self.n_objects)
        self.__check_dataset_validity__()



import copy
from datetime import datetime
import logging
import traceback

from keras import backend as K
from keras.metrics import categorical_accuracy
import numpy as np
from sklearn.model_selection import ShuffleSplit
from sklearn.utils import check_random_state
from skopt import Optimizer
from skopt.space import Categorical
from skopt.space import check_dimension
from skopt.space import Integer
from skopt.utils import cook_estimator
from skopt.utils import dump
from skopt.utils import load
from skopt.utils import normalize_dimensions
import tensorflow as tf

from csrank.constants import CHOICE_FUNCTION
from csrank.constants import DISCRETE_CHOICE
from csrank.constants import DYAD_RANKING
from csrank.constants import LABEL_RANKING
from csrank.constants import OBJECT_RANKING
from csrank.learner import Learner
from csrank.metrics import *
from csrank.metrics_np import *
from csrank.tensorflow_util import get_mean_loss
from csrank.tunable import Tunable
from csrank.util import convert_to_loss
from csrank.util import create_dir_recursively
from csrank.util import duration_till_now
from csrank.util import seconds_to_time

PARAMETER_OPTIMIZER = "ParameterOptimizer"

accuracy_scores = [
    f1_measure,
    precision,
    recall,
    instance_informedness,
    zero_one_accuracy,
    zero_one_accuracy_for_scores_np,
    zero_one_accuracy_for_scores,
    kendalls_tau_for_scores_np,
    spearman_correlation_for_scores_scipy,
    kendalls_tau_for_scores,
    spearman_correlation_for_scores,
    make_ndcg_at_k_loss,
    categorical_accuracy_np,
    topk_categorical_accuracy_np,
    topk_categorical_accuracy,
    categorical_accuracy,
]


class TuningCallback(object):
    def __init__(self):
        pass

    def set_optimizer(self, opt):
        self.optimizer = opt

    def on_optimization_begin(self, logs=None):
        pass

    def on_optimization_end(self, logs=None):
        pass

    def on_iteration_begin(self, i, logs=None):
        pass

    def on_iteration_end(self, i, logs=None):
        pass


class ParameterOptimizer(Learner):
    def __init__(
        self,
        learner,
        optimizer_path,
        tunable_parameter_ranges,
        fit_params=None,
        random_state=None,
        tuning_callbacks=None,
        validation_loss=None,
        learning_problem=OBJECT_RANKING,
        **kwd,
    ):
        """

        Parameters
        ----------
        learner : object
            The ranker object for which the hyper parameters needs to optimized.
        optimizer_path : string
            The path where the complete optimizer pickle object can be stored.
        tunable_parameter_ranges : dict
            Dictionary with keys as the object needs to be tuned. Values is the dictionary of tunable parameters with ranges in which they should be tuned.
        fit_params : dict
            The fitting parameters for the ranker
        random_state : int or object
            Numpy random state
        tuning_callbacks : list of object
            Callbacks for the parameter optimizer
        validation_loss : function
            Differentiable loss function for the ranker
        learning_problem : string
            The learning problem under which the ranker comes
        **kwd :
            Keyword arguments for the hidden units
        """
        self.logger = logging.getLogger(PARAMETER_OPTIMIZER)

        create_dir_recursively(optimizer_path, True)
        self.optimizer_path = optimizer_path

        self._tunable_parameter_ranges = tunable_parameter_ranges

        check_learner_class(learner)
        self.learner = learner

        if tuning_callbacks is None:
            self.tuning_callbacks = []
        else:
            self.tuning_callbacks = tuning_callbacks
        loss_funcs = {
            OBJECT_RANKING: zero_one_rank_loss,
            LABEL_RANKING: zero_one_rank_loss,
            DISCRETE_CHOICE: categorical_accuracy,
            DYAD_RANKING: zero_one_rank_loss,
            CHOICE_FUNCTION: f1_measure,
        }
        if validation_loss is None:
            self.validation_loss = loss_funcs[learning_problem]
            self.logger.info(
                "Loss function is not specified, using {}".format(
                    loss_funcs[learning_problem].__name__
                )
            )
        else:
            self.validation_loss = validation_loss
            self.logger.info(
                "Using Loss function {}".format(self.validation_loss.__name__)
            )

        if self.validation_loss in accuracy_scores:
            self.validation_loss = convert_to_loss(self.validation_loss)
            self.logger.info(
                "Loss function specified is an accuracy converting "
                "{}".format(self.validation_loss.__name__)
            )

        if fit_params is None:
            self._fit_params = {}
            self.logger.warning(
                "Fit params are empty, the default parameters will be applied"
            )
        else:
            self._fit_params = fit_params

        self.random_state = random_state
        self.model = None
        self.opt = None

    def _callbacks_set_optimizer(self, opt):
        for cb in self.tuning_callbacks:
            cb.set_optimizer(opt)

    def _callbacks_on_optimization_begin(self):
        self.logger.info("optimizer begin")
        for cb in self.tuning_callbacks:
            cb.on_optimization_begin()

    def _callbacks_on_optimization_end(self):
        self.logger.info("optimizer end")
        for cb in self.tuning_callbacks:
            cb.on_optimization_end()

    def _callbacks_on_iteration_begin(self, t):
        self.logger.info(
            "********************** optimizer iteration begin ****************"
            "****** " + repr(t)
        )
        for cb in self.tuning_callbacks:
            cb.on_iteration_begin(t)

    def _callbacks_on_iteration_end(self, t):
        self.logger.info("current optimizer iteration end " + repr(t))
        for cb in self.tuning_callbacks:
            cb.on_iteration_end(t)

    @property
    def parameter_ranges(self):
        result = []
        for ranges in self._tunable_parameter_ranges.values():
            result.extend(ranges.values())
        return result

    def _set_new_parameters(self, point):
        i = 0
        # We are iterating over all elements of the dictionary in order.
        # This works because dicts are order-preserving.
        for obj, ranges in self._tunable_parameter_ranges.items():
            param_dict = dict()
            for j, p in enumerate(ranges.keys()):
                param_dict[p] = point[i + j]
            self.logger.info(
                "obj: {}, current parameters {}".format(type(obj).__name__, param_dict)
            )
            obj.set_tunable_parameters(**param_dict)
            i += len(ranges)

    def _fit_ranker(self, xtrain, ytrain, xtest, ytest, next_point):
        start = datetime.now()
        self._set_new_parameters(next_point)
        try:
            self.learner.fit(xtrain, ytrain, **self._fit_params)
            ypred = self.learner(xtest)
            loss = get_mean_loss(self.validation_loss, ytest, ypred)
            time_taken = duration_till_now(start)
        except Exception:
            self.logger.error(traceback.format_exc())
            self.logger.info(
                "For current parameter error occurred so taking loss as maximum value"
            )
            loss = 1.00
            time_taken = duration_till_now(start)
        return loss, time_taken

    def fit(
        self,
        X,
        Y,
        total_duration=600,
        n_iter=100,
        cv_iter=None,
        acq_func="gp_hedge",
        **kwargs,
    ):
        start = datetime.now()
        self.random_state_ = check_random_state(self.random_state)

        def splitter(itr):
            for train_idx, test_idx in itr:
                yield X[train_idx], Y[train_idx], X[test_idx], Y[test_idx]

        def splitter_dict(itr_dict):
            n_splits = len(list(itr_dict.values())[0])
            for i in range(n_splits):
                X_train = dict()
                Y_train = dict()
                X_test = dict()
                Y_test = dict()
                for n_obj, itr in itr_dict.items():
                    train_idx = itr[i][0]
                    test_idx = itr[i][1]
                    X_train[n_obj] = np.copy(X[n_obj][train_idx])
                    X_test[n_obj] = np.copy(X[n_obj][test_idx])
                    Y_train[n_obj] = np.copy(Y[n_obj][train_idx])
                    Y_test[n_obj] = np.copy(Y[n_obj][test_idx])
                yield X_train, Y_train, X_test, Y_test

        if cv_iter is None:
            cv_iter = ShuffleSplit(
                n_splits=3, test_size=0.1, random_state=self.random_state_,
            )
        if isinstance(X, dict):
            splits = dict()
            for n_obj, arr in X.items():
                if arr.shape[0] == 1:
                    splits[n_obj] = [([0], [0]) for i in range(cv_iter.n_splits)]
                else:
                    splits[n_obj] = list(cv_iter.split(arr))
        else:
            splits = list(cv_iter.split(X))
        # Pre-compute splits for reuse
        # Here we fix a random seed for all simulations to correlate the random
        # streams:

        seed = self.random_state_.randint(2 ** 32, dtype="uint32")
        self.logger.debug("Random seed for the ranking algorithm: {}".format(seed))
        opt_seed = self.random_state_.randint(2 ** 32, dtype="uint32")
        self.logger.debug("Random seed for the optimizer: {}".format(opt_seed))
        gp_seed = self.random_state_.randint(2 ** 32, dtype="uint32")
        self.logger.debug("Random seed for the GP surrogate: {}".format(gp_seed))
        n_iter = self.set_optimizer(n_iter, opt_seed, acq_func, gp_seed, **kwargs)

        self._callbacks_set_optimizer(self.opt)
        self._callbacks_on_optimization_begin()
        time_taken = duration_till_now(start)
        total_duration -= time_taken
        max_fit_duration = -np.inf
        self.logger.info(
            "Time left for {} iterations is {}".format(
                n_iter, seconds_to_time(total_duration)
            )
        )

        try:
            for t in range(n_iter):
                if total_duration <= 0:
                    break
                start = datetime.now()
                self._callbacks_on_iteration_begin(t)
                self.logger.info("Starting optimization iteration: {}".format(t))
                if t > 0:
                    self.log_best_params()
                next_point = self.opt.ask()
                self.logger.info("Next parameters:\n{}".format(next_point))
                results = []
                running_times = []
                if isinstance(X, dict):
                    for X_train, Y_train, X_test, Y_test in splitter_dict(splits):
                        result, time_taken = self._fit_ranker(
                            X_train, Y_train, X_test, Y_test, next_point
                        )
                        running_times.append(time_taken)
                        results.append(result)
                else:
                    for X_train, Y_train, X_test, Y_test in splitter(splits):
                        result, time_taken = self._fit_ranker(
                            X_train, Y_train, X_test, Y_test, next_point
                        )
                        running_times.append(time_taken)
                        results.append(result)

                results = np.array(results)
                running_times = np.array(running_times)
                mean_result = np.mean(results)
                mean_fitting_duration = np.mean(running_times)

                # Storing the maximum time to run the splitting model and adding the time for out of sample evaluation
                if max_fit_duration < np.sum(running_times):
                    max_fit_duration = np.sum(running_times)

                self.logger.info(
                    "Validation error for the parameters is {:.4f}".format(mean_result)
                )
                self.logger.info(
                    "Time taken for the parameters is {}".format(
                        seconds_to_time(np.sum(running_times))
                    )
                )
                if "ps" in self.opt.acq_func:
                    self.opt.tell(next_point, [mean_result, mean_fitting_duration])
                else:
                    self.opt.tell(next_point, mean_result)
                self._callbacks_on_iteration_end(t)

                self.logger.info(
                    "Main optimizer iterations done {} and saving the model".format(
                        np.array(self.opt.yi).shape[0]
                    )
                )
                dump(self.opt, self.optimizer_path)

                time_taken = duration_till_now(start)
                total_duration -= time_taken
                self.logger.info(
                    "Time left for simulations is {} ".format(
                        seconds_to_time(total_duration)
                    )
                )

                # Delete Tensorflow graph, to prevent memory leaks:
                K.clear_session()
                sess = tf.Session()
                K.set_session(sess)
                if (total_duration - max_fit_duration) < 0:
                    self.logger.info(
                        "Maximum time required by model to validate parameter values {}".format(
                            seconds_to_time(max_fit_duration)
                        )
                    )
                    self.logger.info(
                        "At iteration {} simulation stops, due to time deficiency".format(
                            t
                        )
                    )
                    break

        except KeyboardInterrupt:
            self.logger.debug(
                "Optimizer interrupted saving the model at {}".format(
                    self.optimizer_path
                )
            )
            self.log_best_params()
        else:
            self.logger.debug(
                "Finally, fit a model on the complete training set and storing the model at {}".format(
                    self.optimizer_path
                )
            )

        finally:
            K.clear_session()
            sess = tf.Session()
            K.set_session(sess)
            self._callbacks_on_optimization_end()
            # self._fit_params["epochs"] = np.min([self._fit_params.get("epochs", 500) * 2, 1000])
            if "ps" in self.opt.acq_func:
                best_point = self.opt.Xi[np.argmin(np.array(self.opt.yi)[:, 0])]
            else:
                best_point = self.opt.Xi[np.argmin(self.opt.yi)]
            self._set_new_parameters(best_point)
            self.model = copy.copy(self.learner)
            self.model.fit(X, Y, **self._fit_params)
            if np.array(self.opt.yi).shape[0] != 0:
                dump(self.opt, self.optimizer_path)

    def log_best_params(self):
        if "ps" in self.opt.acq_func:
            best_i = np.argmin(np.array(self.opt.yi)[:, 0])
            best_loss = self.opt.yi[best_i]
            best_params = self.opt.Xi[best_i]
            self.logger.info(
                "Best parameters so far with a loss of {:.4f}:\n {}".format(
                    best_loss[0], best_params
                )
            )
        else:
            best_i = np.argmin(self.opt.yi)
            best_loss = self.opt.yi[best_i]
            best_params = self.opt.Xi[best_i]
            self.logger.info(
                "Best parameters so far with a loss of {:.4f}:\n {}".format(
                    best_loss, best_params
                )
            )

    def set_optimizer(self, n_iter, opt_seed, acq_func, gp_seed, **kwargs):
        self.logger.info("Retrieving model stored at: {}".format(self.optimizer_path))
        try:
            optimizer = load(self.optimizer_path)
            self.logger.info("Loading model stored at: {}".format(self.optimizer_path))
            finished_iter = np.array(optimizer.yi).shape[0]
            if finished_iter == 0:
                optimizer = None
                self.logger.info(
                    "Optimizer did not finish any iterations so setting optimizer to null"
                )
        except KeyError:
            self.logger.error("Cannot open the file {}".format(self.optimizer_path))
            optimizer = None

        except ValueError:
            self.logger.error("Cannot open the file {}".format(self.optimizer_path))
            optimizer = None
        except FileNotFoundError:
            self.logger.error(
                "No such file or directory: {}".format(self.optimizer_path)
            )
            optimizer = None

        if optimizer is not None:
            n_iter = n_iter - finished_iter
            if n_iter < 0:
                n_iter = 0
            self.logger.info(
                "Iterations already done: {} and running iterations {}".format(
                    finished_iter, n_iter
                )
            )
            self.opt = optimizer
            self.logger.debug("Setting the provided optimizer")
            self.log_best_params()
        else:
            transformed = []
            for param in self.parameter_ranges:
                transformed.append(check_dimension(param))
            self.logger.info("Parameter Space: {}".format(transformed))
            norm_space = normalize_dimensions(transformed)
            self.logger.info(
                "Parameter Space after transformation: {}".format(norm_space)
            )
            categorical_space = np.array(
                [
                    isinstance(s, Categorical) or isinstance(s, Integer)
                    for s in norm_space
                ]
            )
            self.logger.info("categorical_space: {}".format(categorical_space))
            if np.all(categorical_space):
                base_estimator = cook_estimator(
                    "RF", space=norm_space, random_state=gp_seed
                )
            else:
                base_estimator = cook_estimator(
                    "GP", space=norm_space, random_state=gp_seed, noise="gaussian"
                )

            self.opt = Optimizer(
                dimensions=self.parameter_ranges,
                random_state=opt_seed,
                base_estimator=base_estimator,
                acq_func=acq_func,
                **kwargs,
            )

        return n_iter

    def predict_pair(self, a, b, **kwargs):
        if self.model is not None:
            return self.model.predict_pair(a, b, **kwargs)
        else:
            self.logger.error("The ranking model was not fit yet.")
            raise AttributeError

    def _predict_scores_fixed(self, X, **kwargs):
        if self.model is not None:
            return self.model._predict_scores_fixed(X, **kwargs)
        else:
            self.logger.error("The ranking model was not fit yet.")
            raise AttributeError

    def predict_scores(self, X, **kwargs):
        if self.model is not None:
            return self.model.predict_scores(X, **kwargs)
        else:
            self.logger.error("The ranking model was not fit yet.")
            raise AttributeError

    def predict(self, X, **kwargs):
        if self.model is not None:
            return self.model.predict(X, **kwargs)
        else:
            self.logger.error("The ranking model was not fit yet.")
            raise AttributeError

    def predict_for_scores(self, scores, **kwargs):
        if self.model is not None:
            return self.model.predict_for_scores(scores, **kwargs)
        else:
            self.logger.error("The ranking model was not fit yet.")
            raise AttributeError

    def set_tunable_parameters(self, **point):
        self.model.set_tunable_parameters(**point)


def check_learner_class(ranker):
    """ Function which checks if the ranker is an instance of the :class:`csrank.tunning.Tunable` class

    Parameters
    ----------
    ranker: object
        The ranker object to be checked
    """
    if not (isinstance(ranker, Tunable) and hasattr(ranker, "set_tunable_parameters")):
        logging.error("The given object ranker is not tunable")
        raise AttributeError



import logging

import numpy as np
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from sklearn.utils import check_random_state

from csrank.learner import Learner
from csrank.numpy_util import normalize
from csrank.objectranking.object_ranker import ObjectRanker
from csrank.util import print_dictionary
from ..dataset_reader.objectranking.util import complete_linear_regression_dataset

__all__ = ["ExpectedRankRegression"]


class ExpectedRankRegression(ObjectRanker, Learner):
    def __init__(
        self,
        alpha=0.0,
        l1_ratio=0.5,
        tol=1e-4,
        normalize=True,
        fit_intercept=True,
        random_state=None,
        **kwargs,
    ):
        """
            Create an expected rank regression model.
            This model normalizes the ranks to [0, 1] and treats them as regression target. For ``α = 0`` we employ
            simple linear regression. For α > 0 the model becomes ridge regression (when ``l1_ratio = 0``) or
            elastic net (when ``l1_ratio > 0``). The target for an object :math:`x_k \\in Q` is:


            .. math::
                r(x_k) = \\frac{\\pi(k)}{n} \\quad ,

            where :math:`\\pi(k)` is the rank of the :math:`x_k`. The regression model learns a function
            :math:`F \\colon \\mathcal{X} \\to \\mathbb{R}`. The ranking for the given query set :math:`Q` defined as:

            .. math::
                ρ(Q)  = \\operatorname{argsort}_{x \\in Q}  \\; -1*F(x)


            Parameters
            ----------
            alpha : float, optional
                Regularization strength
            l1_ratio : float, optional
                Ratio between pure L2 (=0) or pure L1 (=1) regularization.
            tol : float, optional
                Optimization tolerance
            normalize : bool, optional
                If True, the regressors will be normalized before fitting.
            fit_intercept : bool, optional
                If True, the linear model will also fit an intercept.
            random_state : int, RandomState instance or None, optional
                Seed of the pseudorandom generator or a RandomState instance
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Kamishima, T., Kazawa, H., & Akaho, S. (2005, November). "Supervised ordering-an empirical survey.", Fifth IEEE International Conference on Data Mining.
        """
        self.normalize = normalize
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.tol = tol
        self.logger = logging.getLogger(ExpectedRankRegression.__name__)
        self.fit_intercept = fit_intercept
        self.random_state = random_state
        self.weights = None

    def fit(self, X, Y, **kwargs):
        """
            Fit an ExpectedRankRegression on the provided set of queries X and preferences Y of those objects.
            The provided queries and corresponding preferences are of a fixed size (numpy arrays).

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Rankings of the given objects
            **kwargs
                Keyword arguments for the fit function
        """
        self.random_state_ = check_random_state(self.random_state)
        self.logger.debug("Creating the Dataset")
        x_train, y_train = complete_linear_regression_dataset(X, Y)
        self.logger.debug("Finished the Dataset")
        if self.alpha < 1e-3:
            self.model = LinearRegression(
                normalize=self.normalize, fit_intercept=self.fit_intercept
            )
            self.logger.info("LinearRegression")
        else:
            if self.l1_ratio >= 0.01:
                self.model = ElasticNet(
                    alpha=self.alpha,
                    l1_ratio=self.l1_ratio,
                    normalize=self.normalize,
                    tol=self.tol,
                    fit_intercept=self.fit_intercept,
                    random_state=self.random_state_,
                )
                self.logger.info("Elastic Net")
            else:
                self.model = Ridge(
                    alpha=self.alpha,
                    normalize=self.normalize,
                    tol=self.tol,
                    fit_intercept=self.fit_intercept,
                    random_state=self.random_state_,
                )
                self.logger.info("Ridge")
        self.logger.debug("Finished Creating the model, now fitting started")
        self.model.fit(x_train, y_train)
        self.weights = self.model.coef_.flatten()
        if self.fit_intercept:
            self.weights = np.append(self.weights, self.model.intercept_)
        self.logger.debug("Fitting Complete")

    def _predict_scores_fixed(self, X, **kwargs):
        n_instances, n_objects, n_features = X.shape
        self.logger.info(
            "For Test instances {} objects {} features {}".format(*X.shape)
        )
        X1 = X.reshape(n_instances * n_objects, n_features)
        scores = n_objects - self.model.predict(X1)
        scores = scores.reshape(n_instances, n_objects)
        scores = normalize(scores)
        self.logger.info("Done predicting scores")
        return scores

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ObjectRanker.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def set_tunable_parameters(self, alpha=0.0, l1_ratio=0.5, tol=1e-4, **point):
        """
            Set tunable parameters of the PairwiseSVM model to the values provided.

            Parameters
            ----------
            alpha : float
               Regularization strength
            l1_ratio : float in [0,1]
                Ratio between pure L2 (=0) or pure L1 (=1) regularization
            tol : float
                 Optimization tolerance for regression model
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        self.tol = tol
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support"
                " tunable parameters"
                " called: {}".format(print_dictionary(point))
            )



from abc import ABCMeta

from csrank.constants import DISCRETE_CHOICE
from csrank.dataset_reader.discretechoice.util import convert_to_label_encoding

__all__ = ["DiscreteObjectChooser"]


class DiscreteObjectChooser(metaclass=ABCMeta):
    @property
    def learning_problem(self):
        return DISCRETE_CHOICE

    def predict_for_scores(self, scores):
        """
            Binary discrete choice vector :math:`y` represents the choices amongst the objects in :math:`Q`, such that
            :math:`y(k) = 1` represents that the object :math:`x_k` is chosen and :math:`y(k) = 0` represents
            it is not chosen. For choice to be discrete :math:`\\sum_{x_i \\in Q} y(i) = 1`. Predict discrete choices for
            the scores for a given collection of sets of objects (query sets).

            Parameters
            ----------
            scores : dict or numpy array
                Dictionary with a mapping from query set size to numpy arrays
                or a single numpy array of size containing scores of each object of size:
                (n_instances, n_objects)


            Returns
            -------
            Y : dict or numpy array
                Dictionary with a mapping from query set size to numpy arrays
                or a single numpy array containing predicted discrete choice vectors of size:
                (n_instances, n_objects)
        """
        if isinstance(scores, dict):
            result = dict()
            for n, s in scores.items():
                result[n] = s.argmax(axis=1)
                result[n] = convert_to_label_encoding(result[n], n)

        else:
            n = scores.shape[-1]
            result = scores.argmax(axis=1)
            result = convert_to_label_encoding(result, n)
        return result



import numpy as np

from csrank.constants import CHOICE_FUNCTION
from ..mnist_dataset_reader import MNISTDatasetReader


class MNISTChoiceDatasetReader(MNISTDatasetReader):
    def __init__(self, dataset_type="unique", **kwargs):
        dataset_func_dict = {
            "unique": self.create_dataset_unique,
            "largest": self.create_dataset_largest,
            "mode": self.create_dataset_mode,
        }
        if dataset_type not in dataset_func_dict:
            raise ValueError(
                f"dataset_type must be one of {set(dataset_func_dict.keys())}"
            )
        self.dataset_function = dataset_func_dict[dataset_type]
        super(MNISTChoiceDatasetReader, self).__init__(
            learning_problem=CHOICE_FUNCTION, **kwargs
        )
        self.logger.info("Dataset type {}".format(dataset_type))

    def create_dataset_largest(self):
        self.logger.info("Largest Dataset")
        num_classes = len(np.unique(self.y_labels))
        n_total = self.n_test_instances + self.n_train_instances
        largest_numbers = self.random_state.randint(1, num_classes, size=n_total)
        self.X = np.empty((n_total, self.n_objects, self.n_features))
        y_number = np.empty((n_total, self.n_objects), dtype=int)
        for i in range(n_total):
            remaining = self.X_raw[self.y_labels <= largest_numbers[i]]
            while True:
                indices = self.random_state.choice(
                    len(remaining), size=self.n_objects, replace=False
                )
                self.X[i] = remaining[indices]
                y_number[i] = self.y_labels[self.y_labels <= largest_numbers[i]][
                    indices
                ]
                if largest_numbers[i] in y_number[i]:
                    break
        self.Y = (y_number == largest_numbers[:, None]).astype(int)
        self.__check_dataset_validity__()

    def create_dataset_mode(self):
        self.logger.info("Mode Dataset")
        n_total = self.n_test_instances + self.n_train_instances
        self.X = np.empty((n_total, self.n_objects, self.n_features))
        self.Y = np.zeros((n_total, self.n_objects), dtype=int)
        all_indices = np.arange(len(self.X_raw))
        for i in range(n_total):
            indices = self.random_state.choice(
                all_indices, size=self.n_objects, replace=False
            )
            labels = self.y_labels[indices]
            numbers, counts = np.unique(labels, return_counts=True)
            modes = numbers[np.where(counts == np.max(counts))[0]]
            self.X[i] = self.X_raw[indices]
            self.Y[i] = np.array(np.isin(labels, modes), dtype=int)
        self.__check_dataset_validity__()

    def create_dataset_unique(self):
        self.logger.info("Unique Dataset")
        n_total = self.n_test_instances + self.n_train_instances
        self.X = np.empty((n_total, self.n_objects, self.n_features))
        self.Y = np.zeros((n_total, self.n_objects), dtype=int)
        all_indices = np.arange(len(self.X_raw))
        for i in range(n_total):
            while True:
                indices = self.random_state.choice(
                    all_indices, size=self.n_objects, replace=False
                )
                labels = self.y_labels[indices]
                numbers, counts = np.unique(labels, return_counts=True)
                unique_numbers = numbers[np.where(counts == 1)[0]]
                if len(unique_numbers) > 1:
                    self.random_state.shuffle(indices)
                    self.X[i] = self.X_raw[indices]
                    labels = self.y_labels[indices]
                    self.Y[i] = np.array(np.isin(labels, unique_numbers), dtype=int)
                    break
        self.__check_dataset_validity__()



import logging

from keras.optimizers import SGD
from keras.regularizers import l2
from sklearn.model_selection import train_test_split

from csrank.choicefunction.choice_functions import ChoiceFunctions
from csrank.choicefunction.util import generate_complete_pairwise_dataset
from csrank.core.cmpnet_core import CmpNetCore


class CmpNetChoiceFunction(CmpNetCore, ChoiceFunctions):
    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        loss_function="binary_crossentropy",
        batch_normalization=True,
        kernel_regularizer=l2(1e-4),
        kernel_initializer="lecun_normal",
        activation="relu",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=["binary_accuracy"],
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create an instance of the :class:`CmpNetCore` architecture for learning a choice function.
            CmpNet breaks the preferences in form of rankings into pairwise comparisons and learns a pairwise model for
            the each pair of object in the underlying set. For prediction list of objects is converted in pair of
            objects and the pairwise predicate is evaluated using them. The outputs of the network for each pair of
            objects :math:`U(x_1,x_2), U(x_2,x_1)` are evaluated.
            :math:`U(x_1,x_2)` is a measure of how favorable it is to choose :math:`x_1` over :math:`x_2`.
            The utility score of object :math:`x_i` in query set
            :math:`Q = \\{ x_1 , \\ldots , x_n \\}` is evaluated as:

            .. math::

                U(x_i) = \\left\\{ \\frac{1}{n-1} \\sum_{j \\in [n]
                \\setminus \\{i\\}} U_1(x_i , x_j)\\right\\}

            The choice set is defined as:

            .. math::

                c(Q) = \\{ x_i \\in Q \\lvert \\, U(x_i) > t \\}

            Parameters
            ----------
            n_hidden : int
                Number of hidden layers used in the scoring network
            n_units : int
                Number of hidden units in each layer of the scoring network
            loss_function : function or string
                Loss function to be used for the binary decision task of the pairwise comparisons
            batch_normalization : bool
                Whether to use batch normalization in each hidden layer
            kernel_regularizer : function
                Regularizer function applied to all the hidden weight matrices.
            activation : function or string
                Type of activation function to use in each hidden layer
            optimizer : function or string
                Optimizer to use during stochastic gradient descent
            metrics : list
                List of metrics to evaluate during training (can be non-differentiable)
            batch_size : int
                Batch size to use during training
            random_state : int, RandomState instance or None
                Seed of the pseudorandom generator or a RandomState instance
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Leonardo Rigutini, Tiziano Papini, Marco Maggini, and Franco Scarselli. 2011. SortNet: Learning to Rank by a Neural Preference Function. IEEE Trans. Neural Networks 22, 9 (2011), 1368–1380. https://doi.org/10.1109/TNN.2011.2160875

        """
        super().__init__(
            n_hidden=n_hidden,
            n_units=n_units,
            loss_function=loss_function,
            batch_normalization=batch_normalization,
            kernel_regularizer=kernel_regularizer,
            kernel_initializer=kernel_initializer,
            activation=activation,
            optimizer=optimizer,
            metrics=metrics,
            batch_size=batch_size,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(CmpNetChoiceFunction.__name__)
        self.logger.info("Initializing network")
        self.threshold = 0.5

    def _convert_instances_(self, X, Y):
        self.logger.debug("Creating the Dataset")
        x1, x2, garbage, y_double, garbage = generate_complete_pairwise_dataset(X, Y)
        del garbage
        if x1.shape[0] > self.threshold_instances:
            indices = self.random_state.choice(
                x1.shape[0], self.threshold_instances, replace=False
            )
            x1 = x1[indices, :]
            x2 = x2[indices, :]
            y_double = y_double[indices, :]
        self.logger.debug("Finished the Dataset instances {}".format(x1.shape[0]))
        return x1, x2, y_double

    def construct_model(self):
        return super().construct_model()

    def fit(
        self,
        X,
        Y,
        epochs=10,
        callbacks=None,
        validation_split=0.1,
        tune_size=0.1,
        thin_thresholds=1,
        verbose=0,
        **kwd,
    ):
        """
            Fit a CmptNet model for learning a choice fucntion on the provided set of queries X and preferences Y of
            those objects. The provided queries and corresponding preferences are of a fixed size (numpy arrays). For
            learning this network the binary cross entropy loss function for a pair of objects :math:`x_i, x_j \\in Q`
            is defined as:

            .. math::

                C_{ij} =  -\\tilde{P_{ij}}(0)\\cdot \\log(U(x_i,x_j)) - \\tilde{P_{ij}}(1) \\cdot \\log(U(x_j,x_i)) \\ ,

            where :math:`\\tilde{P_{ij}}` is ground truth probability of the preference of :math:`x_i` over :math:`x_j`.
            :math:`\\tilde{P_{ij}} = (1,0)` if :math:`x_i \\succ x_j` else :math:`\\tilde{P_{ij}} = (0,1)`.

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Preferences in form of Orderings or Choices for given n_objects
            epochs : int
                Number of epochs to run if training for a fixed query size
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float (range : [0,1])
                Percentage of instances to split off to validate on
            tune_size: float (range : [0,1])
                Percentage of instances to split off to tune the threshold for the choice function
            thin_thresholds: int
                The number of instances of scores to skip while tuning the threshold
            verbose : bool
                Print verbose information
            **kwd :
                Keyword arguments for the fit function
        """
        if tune_size > 0:
            X_train, X_val, Y_train, Y_val = train_test_split(
                X, Y, test_size=tune_size, random_state=self.random_state
            )
            try:
                super().fit(
                    X_train,
                    Y_train,
                    epochs,
                    callbacks,
                    validation_split,
                    verbose,
                    **kwd,
                )
            finally:
                self.logger.info(
                    "Fitting utility function finished. Start tuning threshold."
                )
        else:
            super().fit(X, Y, epochs, callbacks, validation_split, verbose, **kwd)
            self.threshold = 0.5

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ChoiceFunctions.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def clear_memory(self, **kwargs):
        super().clear_memory(**kwargs)

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden=n_hidden,
            n_units=n_units,
            reg_strength=reg_strength,
            learning_rate=learning_rate,
            batch_size=batch_size,
            **point,
        )



from itertools import combinations
from itertools import permutations
import logging

from keras import backend as K
from keras import Input
from keras import Model
from keras import optimizers
from keras.layers import add
from keras.layers import concatenate
from keras.layers import Dense
from keras.layers import Lambda
from keras.optimizers import SGD
from keras.regularizers import l2
import numpy as np
from sklearn.utils import check_random_state
import tensorflow as tf

from csrank.constants import allowed_dense_kwargs
from csrank.layers import NormalizedDense
from csrank.learner import Learner
from csrank.losses import hinged_rank_loss
from csrank.util import print_dictionary


class FETANetwork(Learner):
    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        add_zeroth_order_model=False,
        max_number_of_objects=5,
        num_subsample=5,
        loss_function=hinged_rank_loss,
        batch_normalization=False,
        kernel_regularizer=l2(1e-4),
        kernel_initializer="lecun_normal",
        activation="selu",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=None,
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        self.logger = logging.getLogger(FETANetwork.__name__)
        self.random_state = random_state
        self.kernel_regularizer = kernel_regularizer
        self.kernel_initializer = kernel_initializer
        self.batch_normalization = batch_normalization
        self.activation = activation
        self.loss_function = loss_function
        self.metrics = metrics
        self.max_number_of_objects = max_number_of_objects
        self.num_subsample = num_subsample
        self.batch_size = batch_size
        self.hash_file = None
        self.optimizer = optimizers.get(optimizer)
        self._optimizer_config = self.optimizer.get_config()
        self._use_zeroth_model = add_zeroth_order_model
        self.n_hidden = n_hidden
        self.n_units = n_units
        keys = list(kwargs.keys())
        for key in keys:
            if key not in allowed_dense_kwargs:
                del kwargs[key]
        self.kwargs = kwargs
        self._pairwise_model = None
        self.model = None
        self._zero_order_model = None

    @property
    def n_objects(self):
        if self.n_objects_fit_ > self.max_number_of_objects:
            return self.max_number_of_objects
        return self.n_objects_fit_

    def _construct_layers(self, **kwargs):
        self.input_layer = Input(
            shape=(self.n_objects_fit_, self.n_object_features_fit_)
        )
        # Todo: Variable sized input
        # X = Input(shape=(None, n_features))
        self.logger.info("n_hidden {}, n_units {}".format(self.n_hidden, self.n_units))
        if self.batch_normalization:
            if self._use_zeroth_model:
                self.hidden_layers_zeroth = [
                    NormalizedDense(
                        self.n_units, name="hidden_zeroth_{}".format(x), **kwargs
                    )
                    for x in range(self.n_hidden)
                ]
            self.hidden_layers = [
                NormalizedDense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        else:
            if self._use_zeroth_model:
                self.hidden_layers_zeroth = [
                    Dense(self.n_units, name="hidden_zeroth_{}".format(x), **kwargs)
                    for x in range(self.n_hidden)
                ]
            self.hidden_layers = [
                Dense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        assert len(self.hidden_layers) == self.n_hidden
        self.output_node = Dense(
            1, activation="sigmoid", kernel_regularizer=self.kernel_regularizer
        )
        if self._use_zeroth_model:
            self.output_node_zeroth = Dense(
                1, activation="sigmoid", kernel_regularizer=self.kernel_regularizer
            )

    @property
    def zero_order_model(self):
        if self._zero_order_model is None and self._use_zeroth_model:
            self.logger.info("Creating zeroth model")
            inp = Input(shape=(self.n_object_features_fit_,))

            x = inp
            for hidden in self.hidden_layers_zeroth:
                x = hidden(x)
            zeroth_output = self.output_node_zeroth(x)

            self._zero_order_model = Model(inputs=[inp], outputs=zeroth_output)
            self.logger.info("Done creating zeroth model")
        return self._zero_order_model

    @property
    def pairwise_model(self):
        if self._pairwise_model is None:
            self.logger.info("Creating pairwise model")
            x1 = Input(shape=(self.n_object_features_fit_,))
            x2 = Input(shape=(self.n_object_features_fit_,))

            x1x2 = concatenate([x1, x2])
            x2x1 = concatenate([x2, x1])

            for hidden in self.hidden_layers:
                x1x2 = hidden(x1x2)
                x2x1 = hidden(x2x1)

            merged_left = concatenate([x1x2, x2x1])
            merged_right = concatenate([x2x1, x1x2])

            n_g = self.output_node(merged_left)
            n_l = self.output_node(merged_right)

            merged_output = concatenate([n_g, n_l])
            self._pairwise_model = Model(inputs=[x1, x2], outputs=merged_output)
            self.logger.info("Done creating pairwise model")
        return self._pairwise_model

    def _predict_pair(self, a, b, only_pairwise=False, **kwargs):
        # TODO: Is this working correctly?
        pairwise = self.pairwise_model.predict([a, b], **kwargs)
        if not only_pairwise and self._use_zeroth_model:
            utility_a = self.zero_order_model.predict([a])
            utility_b = self.zero_order_model.predict([b])
            return pairwise + (utility_a, utility_b)
        return pairwise

    def _predict_scores_using_pairs(self, X, **kwd):
        n_instances, n_objects, n_features = X.shape
        n2 = n_objects * (n_objects - 1)
        pairs = np.empty((n2, 2, n_features))
        scores = np.zeros((n_instances, n_objects))
        for n in range(n_instances):
            for k, (i, j) in enumerate(permutations(range(n_objects), 2)):
                pairs[k] = (X[n, i], X[n, j])
            result = self._predict_pair(
                pairs[:, 0], pairs[:, 1], only_pairwise=True, **kwd
            )[:, 0]
            scores[n] += result.reshape(n_objects, n_objects - 1).mean(axis=1)
            del result
        del pairs
        if self._use_zeroth_model:
            scores_zero = self.zero_order_model.predict(X.reshape(-1, n_features))
            scores_zero = scores_zero.reshape(n_instances, n_objects)
            scores = scores + scores_zero
        return scores

    def construct_model(self):
        """
            Construct the :math:`1`-st order and :math:`0`-th order models, which are used to approximate the
            :math:`U_1(x, C(x))` and the :math:`U_0(x)` utilities respectively. For each pair of objects in
            :math:`x_i, x_j \\in Q` :math:`U_1(x, C(x))` we construct :class:`CmpNetCore` with weight sharing to
            approximate a pairwise-matrix. A pairwise matrix with index (i,j) corresponds to the :math:`U_1(x_i,x_j)`
            is a measure of how favorable it is to choose :math:`x_i` over :math:`x_j`. Using this matrix we calculate
            the borda score for each object to calculate :math:`U_1(x, C(x))`. For `0`-th order model we construct
            :math:`\\lvert Q \\lvert` sequential networks whose weights are shared to evaluate the :math:`U_0(x)` for
            each object in the query set :math:`Q`. The output mode is using linear activation.

            Returns
            -------
            model: keras :class:`Model`
                Neural network to learn the FETA utility score
        """

        def create_input_lambda(i):
            return Lambda(lambda x: x[:, i])

        if self._use_zeroth_model:
            self.logger.debug("Create 0th order model")
            zeroth_order_outputs = []
            inputs = []
            for i in range(self.n_objects_fit_):
                x = create_input_lambda(i)(self.input_layer)
                inputs.append(x)
                for hidden in self.hidden_layers_zeroth:
                    x = hidden(x)
                zeroth_order_outputs.append(self.output_node_zeroth(x))
            zeroth_order_scores = concatenate(zeroth_order_outputs)
            self.logger.debug("0th order model finished")
        self.logger.debug("Create 1st order model")
        outputs = [list() for _ in range(self.n_objects_fit_)]
        for i, j in combinations(range(self.n_objects_fit_), 2):
            if self._use_zeroth_model:
                x1 = inputs[i]
                x2 = inputs[j]
            else:
                x1 = create_input_lambda(i)(self.input_layer)
                x2 = create_input_lambda(j)(self.input_layer)
            x1x2 = concatenate([x1, x2])
            x2x1 = concatenate([x2, x1])

            for hidden in self.hidden_layers:
                x1x2 = hidden(x1x2)
                x2x1 = hidden(x2x1)

            merged_left = concatenate([x1x2, x2x1])
            merged_right = concatenate([x2x1, x1x2])

            n_g = self.output_node(merged_left)
            n_l = self.output_node(merged_right)

            outputs[i].append(n_g)
            outputs[j].append(n_l)
        # convert rows of pairwise matrix to keras layers:
        outputs = [concatenate(x) for x in outputs]

        # compute utility scores:
        scores = [
            Lambda(lambda s: K.mean(s, axis=1, keepdims=True))(x) for x in outputs
        ]
        scores = concatenate(scores)
        self.logger.debug("1st order model finished")
        if self._use_zeroth_model:
            scores = add([scores, zeroth_order_scores])
        model = Model(inputs=self.input_layer, outputs=scores)
        self.logger.debug("Compiling complete model...")
        model.compile(
            loss=self.loss_function, optimizer=self.optimizer, metrics=self.metrics
        )
        return model

    def fit(
        self, X, Y, epochs=10, callbacks=None, validation_split=0.1, verbose=0, **kwd
    ):
        """
            Fit a generic preference learning model on a provided set of queries.
            The provided queries can be of a fixed size (numpy arrays).

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Preferences in form of rankings or choices for given objects
            epochs : int
                Number of epochs to run if training for a fixed query size
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float (range : [0,1])
                Percentage of instances to split off to validate on
            verbose : bool
                Print verbose information
            **kwd :
                Keyword arguments for the fit function
        """
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        self._construct_layers(
            kernel_regularizer=self.kernel_regularizer,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation,
            **self.kwargs,
        )

        self.logger.debug("Enter fit function...")
        self.random_state_ = check_random_state(self.random_state)

        X, Y = self.sub_sampling(X, Y)
        self.model = self.construct_model()
        self.logger.debug("Starting gradient descent...")

        self.model.fit(
            x=X,
            y=Y,
            batch_size=self.batch_size,
            epochs=epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            verbose=verbose,
            **kwd,
        )
        if self.hash_file is not None:
            self.model.save_weights(self.hash_file)

    def sub_sampling(self, X, Y):
        if self.n_objects_fit_ > self.max_number_of_objects:
            bucket_size = int(self.n_objects_fit_ / self.max_number_of_objects)
            idx = self.random_state_.randint(
                bucket_size, size=(len(X), self.n_objects_fit_)
            )
            # TODO: subsampling multiple rankings
            idx += np.arange(start=0, stop=self.n_objects_fit_, step=bucket_size)[
                : self.n_objects_fit_
            ]
            X = X[np.arange(X.shape[0])[:, None], idx]
            Y = Y[np.arange(X.shape[0])[:, None], idx]
            tmp_sort = Y.argsort(axis=-1)
            Y = np.empty_like(Y)
            Y[np.arange(len(X))[:, None], tmp_sort] = np.arange(self.n_objects_fit_)
        return X, Y

    def _predict_scores_fixed(self, X, **kwargs):
        n_objects = X.shape[-2]
        self.logger.info(
            "For Test instances {} objects {} features {}".format(*X.shape)
        )
        if self.n_objects_fit_ != n_objects:
            scores = self._predict_scores_using_pairs(X, **kwargs)
        else:
            scores = self.model.predict(X, **kwargs)
        self.logger.info("Done predicting scores")
        return scores

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        """
            Set tunable parameters of the FETA-network to the values provided.

            Parameters
            ----------
            n_hidden: int
                Number of hidden layers used in the scoring network
            n_units: int
                Number of hidden units in each layer of the scoring network
            reg_strength: float
                Regularization strength of the regularizer function applied to the `kernel` weights matrix
            learning_rate: float
                Learning rate of the stochastic gradient descent algorithm used by the network
            batch_size: int
                Batch size to use during training
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        self.n_hidden = n_hidden
        self.n_units = n_units
        self.kernel_regularizer = l2(reg_strength)
        self.batch_size = batch_size
        self.optimizer = self.optimizer.from_config(self._optimizer_config)
        K.set_value(self.optimizer.lr, learning_rate)
        self._pairwise_model = None
        self._zero_order_model = None
        self._construct_layers(
            kernel_regularizer=self.kernel_regularizer,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation,
            **self.kwargs,
        )
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support tunable parameters"
                " called: {}".format(print_dictionary(point))
            )

    def clear_memory(self, **kwargs):
        """
            Clear the memory, restores the currently fitted model back to prevent memory leaks.

            Parameters
            ----------
            **kwargs :
                Keyword arguments for the function
        """
        if self.hash_file is not None:
            self.model.save_weights(self.hash_file)
            K.clear_session()
            sess = tf.Session()
            K.set_session(sess)

            self._pairwise_model = None
            self._zero_order_model = None
            self.optimizer = self.optimizer.from_config(self._optimizer_config)
            self._construct_layers(
                kernel_regularizer=self.kernel_regularizer,
                kernel_initializer=self.kernel_initializer,
                activation=self.activation,
                **self.kwargs,
            )
            self.model = self.construct_model()
            self.model.load_weights(self.hash_file)
        else:
            self.logger.info("Cannot clear the memory")



import numpy as np
from sklearn.utils import check_random_state

from csrank.constants import DISCRETE_CHOICE
from csrank.dataset_reader.discretechoice.util import convert_to_label_encoding
from csrank.dataset_reader.tag_genome_reader import critique_dist
from csrank.dataset_reader.util import get_key_for_indices
from ..tag_genome_reader import TagGenomeDatasetReader


class TagGenomeDiscreteChoiceDatasetReader(TagGenomeDatasetReader):
    def __init__(self, dataset_type="similarity", **kwargs):
        super(TagGenomeDiscreteChoiceDatasetReader, self).__init__(
            learning_problem=DISCRETE_CHOICE, **kwargs
        )
        dataset_func_dict = {
            "nearest_neighbour": self.make_nearest_neighbour_dataset,
            "critique_fit_less": self.make_critique_fit_dataset(direction=-1),
            "critique_fit_more": self.make_critique_fit_dataset(direction=1),
            "dissimilar_nearest_neighbour": self.make_dissimilar_nearest_neighbour_dataset,
            "dissimilar_critique_more": self.make_dissimilar_critique_dataset(
                direction=1
            ),
            "dissimilar_critique_less": self.make_dissimilar_critique_dataset(
                direction=-1
            ),
        }
        if dataset_type not in dataset_func_dict:
            raise ValueError(
                f"dataset_type must be one of {set(dataset_func_dict.keys())}"
            )
        self.logger.info("Dataset type: {}".format(dataset_type))
        self.dataset_function = dataset_func_dict[dataset_type]

    def make_nearest_neighbour_dataset(self, n_instances, n_objects, seed, **kwargs):
        X, scores = super(
            TagGenomeDiscreteChoiceDatasetReader, self
        ).make_nearest_neighbour_dataset(
            n_instances=n_instances, n_objects=n_objects, seed=seed
        )
        # Higher the similarity lower the rank of the object, getting the object with second highest similarity
        Y = np.argsort(scores, axis=1)[:, -2]
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y

    def make_critique_fit_dataset(self, direction):
        def dataset_generator(n_instances, n_objects, seed, **kwargs):
            X, scores = super(
                TagGenomeDiscreteChoiceDatasetReader, self
            ).make_critique_fit_dataset(
                n_instances=n_instances,
                n_objects=n_objects,
                seed=seed,
                direction=direction,
            )
            Y = scores.argmax(axis=1)
            Y = convert_to_label_encoding(Y, n_objects)
            return X, Y

        return dataset_generator

    def make_dissimilar_nearest_neighbour_dataset(
        self, n_instances, n_objects, seed, **kwargs
    ):
        self.logger.info(
            "For instances {} objects {}, seed {}".format(n_instances, n_objects, seed)
        )
        X, scores = super(
            TagGenomeDiscreteChoiceDatasetReader, self
        ).make_nearest_neighbour_dataset(
            n_instances=n_instances, n_objects=n_objects, seed=seed
        )
        # Higher the similarity lower the rank of the object, getting the object with second highest similarity
        Y = np.argsort(scores, axis=1)[:, 0]
        Y = convert_to_label_encoding(Y, n_objects)
        return X, Y

    def make_dissimilar_critique_dataset(self, direction):
        def dataset_generator(n_instances, n_objects, seed, **kwargs):
            self.logger.info(
                "For instances {} objects {}, seed {}, direction {}".format(
                    n_instances, n_objects, seed, direction
                )
            )
            random_state = check_random_state(seed)
            X = []
            scores = []
            length = int(n_instances / self.n_movies) + 1
            popular_tags = self.get_genre_tag_id()
            for i, feature in enumerate(self.movie_features):
                if direction == 1:
                    quartile_tags = np.where(
                        np.logical_and(feature >= 1 / 3, feature < 2 / 3)
                    )[0]
                else:
                    quartile_tags = np.where(feature > 1 / 2)[0]
                if len(quartile_tags) < length:
                    quartile_tags = popular_tags
                tag_ids = random_state.choice(quartile_tags, size=length)
                distances = [
                    self.similarity_matrix[get_key_for_indices(i, j)]
                    for j in range(self.n_movies)
                ]
                critique_d = critique_dist(
                    feature,
                    self.movie_features,
                    tag_ids,
                    direction=direction,
                    relu=False,
                )
                critique_fit = np.multiply(critique_d, distances)
                orderings = np.argsort(critique_fit, axis=-1)[:, ::-1]
                minimum = np.zeros(length, dtype=int)
                for k, dist in enumerate(critique_fit):
                    quartile = np.percentile(dist, [0, 5])
                    last = np.where(
                        np.logical_and((dist >= quartile[0]), (dist <= quartile[1]))
                    )[0]
                    if i in last:
                        index = np.where(last == i)[0][0]
                        last = np.delete(last, index)
                    minimum[k] = random_state.choice(last, size=1)[0]
                orderings = orderings[:, 0 : n_objects - 2]
                orderings = np.append(orderings, minimum[:, None], axis=1)
                orderings = np.append(
                    orderings, np.zeros(length, dtype=int)[:, None] + i, axis=1
                )
                for o in orderings:
                    random_state.shuffle(o)
                scores.extend(critique_fit[np.arange(length)[:, None], orderings])
                X.extend(self.movie_features[orderings])
            X = np.array(X)
            scores = np.array(scores)
            indices = random_state.choice(X.shape[0], n_instances, replace=False)
            X = X[indices, :, :]
            scores = scores[indices, :]

            Y = scores.argmin(axis=1)
            Y = convert_to_label_encoding(Y, n_objects)
            return X, Y

        return dataset_generator



import logging

from keras.optimizers import SGD
from keras.regularizers import l2

from csrank.core.feta_network import FETANetwork
from csrank.losses import hinged_rank_loss
from .object_ranker import ObjectRanker

__all__ = ["FETAObjectRanker"]


class FETAObjectRanker(FETANetwork, ObjectRanker):
    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        add_zeroth_order_model=False,
        max_number_of_objects=5,
        num_subsample=5,
        loss_function=hinged_rank_loss,
        batch_normalization=False,
        kernel_regularizer=l2(1e-4),
        kernel_initializer="lecun_normal",
        activation="selu",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=None,
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create a FETA-network architecture for object ranking. The first-evaluate-then-aggregate approach
            approximates the context-dependent utility function using the first-order utility function
            :math:`U_1 \\colon \\mathcal{X} \\times \\mathcal{X} \\rightarrow [0,1]` and zeroth-order utility
            function :math:`U_0 \\colon \\mathcal{X} \\rightarrow [0,1]`.
            The scores each object :math:`x` using a context-dependent utility function :math:`U (x, C_i)`:

            .. math::
                 U(x_i, C_i) = U_0(x_i) + \\frac{1}{n-1} \\sum_{x_j \\in Q \\setminus \\{x_i\\}} U_1(x_i , x_j) \\, .

            Training and prediction complexity is quadratic in the number of objects.
            The ranking for the given query set :math:`Q` is defined as:

            .. math::
                ρ(Q)  = \\operatorname{argsort}_{x_i \\in Q}  \\; U (x_i, C_i)

            Parameters
            ----------
            n_hidden : int
                Number of hidden layers
            n_units : int
                Number of hidden units in each layer
            add_zeroth_order_model : bool
                True if the model should include a latent utility function
            max_number_of_objects : int
                The maximum number of objects to train from
            num_subsample : int
                Number of objects to subsample to
            loss_function : function
                Differentiable loss function for the score vector
            batch_normalization : bool
                Whether to use batch normalization in the hidden layers
            kernel_regularizer : function
                Regularizer to use in the hidden units
            kernel_initializer : function or string
                Initialization function for the weights of each hidden layer
            activation : string or function
                Activation function to use in the hidden units
            optimizer : string or function
                Stochastic gradient optimizer
            metrics : list
                List of evaluation metrics (can be non-differentiable)
            batch_size : int
                Batch size to use for training
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the hidden units
        """
        super().__init__(
            n_hidden=n_hidden,
            n_units=n_units,
            add_zeroth_order_model=add_zeroth_order_model,
            max_number_of_objects=max_number_of_objects,
            num_subsample=num_subsample,
            loss_function=loss_function,
            batch_normalization=batch_normalization,
            kernel_regularizer=kernel_regularizer,
            kernel_initializer=kernel_initializer,
            activation=activation,
            optimizer=optimizer,
            metrics=metrics,
            batch_size=batch_size,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(FETAObjectRanker.__name__)

    def construct_model(self):
        return super().construct_model()

    def fit(
        self, X, Y, epochs=10, callbacks=None, validation_split=0.1, verbose=0, **kwd
    ):
        """
            Fit an object ranking learning model on a provided set of queries.
            The provided queries are of a fixed size (numpy arrays).

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Rankings of the given objects
            epochs : int
                Number of epochs to run if training for a fixed query size
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float
                Percentage of instances to split off to validate on
            verbose : bool
                Print verbose information
            **kwd
                Keyword arguments for the fit function
        """
        super().fit(
            X,
            Y,
            epochs=epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            verbose=verbose,
            **kwd,
        )

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ObjectRanker.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def clear_memory(self, **kwargs):
        self.logger.info("Clearing memory")
        super().clear_memory(**kwargs)

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden=n_hidden,
            n_units=n_units,
            reg_strength=reg_strength,
            learning_rate=learning_rate,
            batch_size=batch_size,
            **point,
        )



from abc import ABCMeta
import collections
import glob
import logging
import os

import numpy as np

from csrank.dataset_reader.util import standardize_features
from csrank.util import create_dir_recursively
from csrank.util import print_dictionary
from .dataset_reader import DatasetReader

try:
    import h5py
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("h5py", "data")


class LetorRankingDatasetReader(DatasetReader, metaclass=ABCMeta):
    def __init__(self, year=2007, fold_id=0, exclude_qf=False, **kwargs):
        super(LetorRankingDatasetReader, self).__init__(
            dataset_folder="letor", **kwargs
        )
        self.DATASET_FOLDER_2007 = "MQ{}".format(year)
        self.DATASET_FOLDER_2008 = "MQ{}".format(year)
        self.logger = logging.getLogger(LetorRankingDatasetReader.__name__)

        if year not in [2007, 2008]:
            raise ValueError("year must be either 2007 or 2008")
        self.year = year
        self.exclude_qf = exclude_qf
        self.query_feature_indices = [4, 5, 6, 19, 20, 21, 34, 35, 36]
        self.query_document_feature_indices = np.delete(
            np.arange(0, 46), self.query_feature_indices
        )
        self.logger.info(
            "For Year {} excluding query features {}".format(self.year, self.exclude_qf)
        )

        self.dataset_indices = np.arange(5)
        self.condition = np.zeros(5, dtype=bool)
        self.file_format = os.path.join(self.dirname, str(self.year), "S{}.h5")
        create_dir_recursively(self.file_format, is_file_path=True)

        for i in self.dataset_indices:
            h5py_file_pth = self.file_format.format(i + 1)
            self.condition[i] = os.path.isfile(h5py_file_pth)
            if not os.path.isfile(h5py_file_pth):
                self.logger.info("File {} not created".format(h5py_file_pth))
        assert (
            fold_id in self.dataset_indices
        ), "For fold {} no test dataset present".format(fold_id + 1)
        self.logger.info("Test dataset is S{}".format(fold_id + 1))
        self.fold_id = fold_id

    def __load_dataset__(self):
        if not (self.condition.all()):
            self.logger.info("HDF5 datasets not created.....")
            self.logger.info("Query features {}".format(self.query_feature_indices))
            self.logger.info(
                "Query Document features {}".format(self.query_document_feature_indices)
            )
            if self.year == "2007":
                mq_2007_files = glob.glob(
                    os.path.join(self.dirname, self.DATASET_FOLDER_2007, "S*.txt")
                )
                self.dataset_dictionaries = self.create_dataset_dictionary(
                    mq_2007_files
                )
            else:
                mq_2008_files = glob.glob(
                    os.path.join(self.dirname, self.DATASET_FOLDER_2008, "S*.txt")
                )
                self.dataset_dictionaries = self.create_dataset_dictionary(
                    mq_2008_files
                )
            for key, dataset in self.dataset_dictionaries.items():
                hdf5file_path = self.file_format.format(key.split("S")[-1])
                self.create_choices_dataset(dataset, hdf5file_path)
        self.X_train = dict()
        self.Y_train = dict()
        for i in self.dataset_indices:
            h5py_file_path = self.file_format.format(i + 1)
            if i != self.fold_id:
                X, Y = self.get_choices_dict(h5py_file_path)
                self.merge_to_train(X, Y)
            else:
                self.X_test, self.Y_test = self.get_choices_dict(h5py_file_path)
        self.logger.info("Done loading the dataset")

    def create_choices_dataset(self, dataset, hdf5file_path):
        self.logger.info("Writing in hd5 {}".format(hdf5file_path))
        X, scores = self.create_instances(dataset)
        result, freq = self._build_training_buckets(X, scores)
        h5f = h5py.File(hdf5file_path, "w")
        self.logger.info("Frequencies of rankings: {}".format(print_dictionary(freq)))

        for key, value in result.items():
            x, s = value
            h5f.create_dataset(
                "X_" + str(key), data=x, compression="gzip", compression_opts=9
            )
            h5f.create_dataset(
                "score_" + str(key), data=s, compression="gzip", compression_opts=9
            )
        lengths = np.array(list(result.keys()))
        h5f.create_dataset(
            "lengths", data=lengths, compression="gzip", compression_opts=9
        )
        h5f.close()

    def get_choices_dict(self, h5py_file_path):
        file = h5py.File(h5py_file_path, "r")
        lengths = file["lengths"]
        X = dict()
        Y = dict()
        for ranking_length in np.array(lengths):
            self.X = np.array(file["X_{}".format(ranking_length)])
            if self.exclude_qf:
                self.X = self.X[:, :, self.query_document_feature_indices]
            self.Y = np.array(file["score_{}".format(ranking_length)])
            self.Y[self.Y == 2] = 1
            self.__check_dataset_validity__()
            X[ranking_length], Y[ranking_length] = self.X, self.Y
        file.close()
        return X, Y

    def merge_to_train(self, X, Y):
        for key in X.keys():
            x = X[key]
            y = Y[key]
            if key in self.X_train:
                self.X_train[key] = np.append(self.X_train[key], x, axis=0)
                self.Y_train[key] = np.append(self.Y_train[key], y, axis=0)
            else:
                self.X_train[key] = x
                self.Y_train[key] = y

    def sub_sampling_from_dictionary(self, train_test="train"):
        X = []
        Y = []
        if train_test == "train":
            Xt = self.X_train
            Yt = self.Y_train
        elif train_test == "test":
            Xt = self.X_test
            Yt = self.Y_test
        for n in Xt.keys():
            if n >= self.n_objects:
                x, y = self.sub_sampling_function(Xt[n], Yt[n])
                if len(x) != 0:
                    if len(X) == 0:
                        X = np.copy(x)
                        Y = np.copy(y)
                    else:
                        X = np.concatenate([X, x], axis=0)
                        Y = np.concatenate([Y, y], axis=0)
        self.logger.info(
            "Sampled instances {} objects {}".format(X.shape[0], X.shape[1])
        )
        return X, Y

    def create_instances(self, dataset):
        X = []
        rel_scores = []
        for k, v in dataset.items():
            x = np.array(v)
            s = x[:, -1]
            f = x[:, 0:-1]
            if not np.all(s == 0):
                rel_scores.append(s)
                X.append(f)
        X = np.array(X)
        rel_scores = np.array(rel_scores)
        return X, rel_scores

    def _build_training_buckets(self, X, scores):
        """Separates object ranking data into buckets of the same ranking size."""
        result = dict()
        frequencies = dict()

        for x, s in zip(X, scores):
            n_objects = len(x)
            if n_objects not in result:
                result[n_objects] = ([], [])
            bucket = result[n_objects]
            bucket[0].append(x)
            bucket[1].append(s)
            if n_objects not in frequencies:
                frequencies[n_objects] = 1
            else:
                frequencies[n_objects] += 1
        # Convert all buckets to numpy arrays:
        for k, v in result.items():
            result[k] = (np.array(v[0]), np.array(v[1]))
        result = collections.OrderedDict(sorted(result.items()))
        return result, frequencies

    def create_dataset_dictionary(self, files):
        self.logger.info("Files {}".format(files))
        dataset_dictionaries = dict()
        for file in files:
            dataset = dict()
            key = os.path.basename(file).split(".txt")[0]
            self.logger.info("File name {}".format(key))
            for line in open(file):
                information = line.split("#")[0].split(" qid:")
                rel_deg = int(information[0])
                qid = information[1].split(" ")[0]
                x = np.array(
                    [float(l.split(":")[1]) for l in information[1].split(" ")[1:-1]]
                )
                x = np.insert(x, len(x), rel_deg)
                if qid not in dataset:
                    dataset[qid] = [x]
                else:
                    dataset[qid].append(x)
            array = np.array([len(i) for i in dataset.values()])
            dataset_dictionaries[key] = dataset
            self.logger.info("Maximum length of ranking: {}".format(np.max(array)))
        return dataset_dictionaries

    def sub_sampling_function(self, Xt, Yt):
        pass

    def get_dataset_dictionaries(self):
        self.X_test, self.X_test = standardize_features(self.X, self.X_test)
        return self.X_train, self.Y_train, self.X_test, self.Y_test

    def get_single_train_test_split(self):
        self.X, self.Y = self.sub_sampling_from_dictionary(train_test="train")
        self.__check_dataset_validity__()
        self.X, self.X_test = standardize_features(self.X, self.X_test)
        return self.X, self.Y, self.X_test, self.Y_test



import logging

from keras.optimizers import SGD
from keras.regularizers import l2

from csrank.core.fate_network import FATENetwork
from csrank.losses import hinged_rank_loss
from csrank.metrics import zero_one_rank_loss_for_scores_ties
from csrank.objectranking.object_ranker import ObjectRanker


class FATEObjectRanker(FATENetwork, ObjectRanker):
    def __init__(
        self,
        n_hidden_set_layers=2,
        n_hidden_set_units=2,
        n_hidden_joint_layers=32,
        n_hidden_joint_units=32,
        activation="selu",
        kernel_initializer="lecun_normal",
        kernel_regularizer=l2(0.01),
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        batch_size=256,
        loss_function=hinged_rank_loss,
        metrics=[zero_one_rank_loss_for_scores_ties],
        random_state=None,
        **kwargs,
    ):
        """
            Create a FATE-network architecture for leaning object ranking function. The first-aggregate-then-evaluate
            approach learns an embedding of each object and then aggregates that into a context representation
            :math:`\\mu_{C(x)}` and then scores each object :math:`x` using a context-dependent utility function
            :math:`U (x, \\mu_{C(x)})`.
            To make it computationally efficient we take the the context :math:`C(x)` as query set :math:`Q`.
            The context-representation is evaluated as:

            .. math::
                \\mu_{C(x)} = \\frac{1}{\\lvert C(x) \\lvert} \\sum_{y \\in C(x)} \\phi(y)

            where :math:`\\phi \\colon \\mathcal{X} \\to \\mathcal{Z}` maps each object :math:`y` to an
            :math:`m`-dimensional embedding space :math:`\\mathcal{Z} \\subseteq \\mathbb{R}^m`.
            Training complexity is quadratic in the number of objects and prediction complexity is only linear.
            The ranking for the given query set :math:`Q` is defined as:

            .. math::
                ρ(Q)  = \\operatorname{argsort}_{x \\in Q}  \\; U (x, \\mu_{C(x)})

            Parameters
            ----------
            n_hidden_set_layers : int
                Number of set layers.
            n_hidden_set_units : int
                Number of hidden set units.
            n_hidden_joint_layers : int
                Number of joint layers.
            n_hidden_joint_units : int
                Number of joint units.
            activation : string or function
                Activation function to use in the hidden units
            kernel_initializer : function or string
                Initialization function for the weights of each hidden layer
            kernel_regularizer : function or string
                Regularizer to use in the hidden units
            optimizer : string or function
                Stochastic gradient optimizer
            batch_size : int
                Batch size to use for training
            loss_function : function
                Differentiable loss function for the score vector
            metrics : list
                List of evaluation metrics (can be non-differentiable)
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the @FATENetwork
        """
        self.loss_function = loss_function
        self.metrics = metrics
        super().__init__(
            n_hidden_set_layers=n_hidden_set_layers,
            n_hidden_set_units=n_hidden_set_units,
            n_hidden_joint_layers=n_hidden_joint_layers,
            n_hidden_joint_units=n_hidden_joint_units,
            activation=activation,
            kernel_initializer=kernel_initializer,
            kernel_regularizer=kernel_regularizer,
            optimizer=optimizer,
            batch_size=batch_size,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(FATEObjectRanker.__name__)

    def construct_model(self, n_features, n_objects):
        return super().construct_model(n_features, n_objects)

    def fit(self, X, Y, **kwd):
        super().fit(X, Y, **kwd)

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ObjectRanker.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def clear_memory(self, **kwargs):
        self.logger.info("Clearing memory")
        super().clear_memory(**kwargs)

    def set_tunable_parameters(
        self,
        n_hidden_set_units=32,
        n_hidden_set_layers=2,
        n_hidden_joint_units=32,
        n_hidden_joint_layers=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden_set_units=n_hidden_set_units,
            n_hidden_set_layers=n_hidden_set_layers,
            n_hidden_joint_units=n_hidden_joint_units,
            n_hidden_joint_layers=n_hidden_joint_layers,
            reg_strength=reg_strength,
            learning_rate=learning_rate,
            batch_size=batch_size,
            **point,
        )



import logging

from keras.optimizers import SGD
from keras.regularizers import l2

from csrank.choicefunction.util import generate_complete_pairwise_dataset
from csrank.core.ranknet_core import RankNetCore
from .discrete_choice import DiscreteObjectChooser


class RankNetDiscreteChoiceFunction(RankNetCore, DiscreteObjectChooser):
    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        loss_function="binary_crossentropy",
        batch_normalization=True,
        kernel_regularizer=l2(1e-4),
        kernel_initializer="lecun_normal",
        activation="relu",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=["binary_accuracy"],
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create an instance of the :class:`RankNetCore` architecture for learning a choice function.
            It breaks the preferences into pairwise comparisons and learns a latent utility model for the objects.
            This network learns a latent utility score for each object in the given query set
            :math:`Q = \\{x_1, \\ldots ,x_n\\}` using the equation :math:`U(x) = F(x, w)` where :math:`w` is the weight
            vector. It is estimated using *pairwise preferences* generated from the discrete choices.

            The discrete choice for the given query set :math:`Q` is defined as:

            .. math::

                ρ(Q)  = \\operatorname{argsort}_{x \\in Q}  \\; U(x)

            Parameters
            ----------
            n_hidden : int
                Number of hidden layers used in the scoring network
            n_units : int
                Number of hidden units in each layer of the scoring network
            loss_function : function or string
                Loss function to be used for the binary decision task of the pairwise comparisons
            batch_normalization : bool
                Whether to use batch normalization in each hidden layer
            kernel_regularizer : function
                Regularizer function applied to all the hidden weight matrices.
            kernel_initializer : function or string
                Initialization function for the weights of each hidden layer
            activation : function or string
                Type of activation function to use in each hidden layer
            optimizer : function or string
                Optimizer to use during stochastic gradient descent
            metrics : list
                List of metrics to evaluate during training (can be non-differentiable)
            batch_size : int
                Batch size to use during training
            random_state : int, RandomState instance or None
                Seed of the pseudo-random generator or a RandomState instance
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Burges, C. et al. (2005, August). "Learning to rank using gradient descent.", In Proceedings of the 22nd international conference on Machine learning (pp. 89-96). ACM.

                [2] Burges, C. J. (2010). "From ranknet to lambdarank to lambdamart: An overview.", Learning, 11(23-581).
        """
        super().__init__(
            n_hidden=n_hidden,
            n_units=n_units,
            loss_function=loss_function,
            batch_normalization=batch_normalization,
            kernel_regularizer=kernel_regularizer,
            kernel_initializer=kernel_initializer,
            activation=activation,
            optimizer=optimizer,
            metrics=metrics,
            batch_size=batch_size,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(RankNetDiscreteChoiceFunction.__name__)
        self.logger.info("Initializing network")

    def construct_model(self):
        return super().construct_model()

    def _convert_instances_(self, X, Y):
        self.logger.debug("Creating the Dataset")
        x1, x2, garbage, garbage, y_single = generate_complete_pairwise_dataset(X, Y)
        del garbage
        if x1.shape[0] > self.threshold_instances:
            indices = self.random_state.choice(
                x1.shape[0], self.threshold_instances, replace=False
            )
            x1 = x1[indices, :]
            x2 = x2[indices, :]
            y_single = y_single[indices]
        self.logger.debug("Finished the Dataset instances {}".format(x1.shape[0]))
        return x1, x2, y_single

    def fit(self, X, Y, **kwd):
        super().fit(X, Y, **kwd)

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return DiscreteObjectChooser.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def clear_memory(self, **kwargs):
        self.logger.info("Clearing memory")
        super().clear_memory(**kwargs)

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden=n_hidden,
            n_units=n_units,
            reg_strength=reg_strength,
            learning_rate=learning_rate,
            batch_size=batch_size,
            **point,
        )



from csrank.core.fate_network import FATENetwork
from csrank.dyadranking.dyad_ranker import DyadRanker
from csrank.numpy_util import scores_to_rankings


class FATEDyadRanker(FATENetwork, DyadRanker):
    def fit(self, Xo, Xc, Y, **kwargs):
        pass

    def predict_scores(self, Xo, Xc, **kwargs):
        return self.model.predict([Xo, Xc], **kwargs)

    def predict(self, Xo, Xc, **kwargs):
        s = self.predict_scores(Xo, Xc, **kwargs)
        return scores_to_rankings(s)



import logging

from keras.optimizers import SGD
from keras.regularizers import l2

from csrank.core.cmpnet_core import CmpNetCore
from csrank.dataset_reader.objectranking.util import generate_complete_pairwise_dataset
from csrank.objectranking.object_ranker import ObjectRanker

__all__ = ["CmpNet"]


class CmpNet(CmpNetCore, ObjectRanker):
    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        loss_function="binary_crossentropy",
        batch_normalization=True,
        kernel_regularizer=l2(1e-4),
        kernel_initializer="lecun_normal",
        activation="relu",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=["binary_accuracy"],
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create an instance of the :class:`CmpNetCore` architecture for learning a object ranking function.
            CmpNet breaks the preferences in form of rankings into pairwise comparisons and learns a pairwise model for
            the each pair of object in the underlying set. For prediction list of objects is converted in pair of
            objects and the pairwise predicate is evaluated using them. The outputs of the network, i.e.,
            :math:`U(x_1,x_2), U(x_2,x_1)`for each pair of objects are evaluated.
            :math:`U(x_1,x_2)` is a measure of how favorable it is to choose :math:`x_1` over :math:`x_2`.
            The utility score of object :math:`x_i` in query set :math:`Q = \\{ x_1 , \\ldots , x_n \\}` is evaluated as:

            .. math::

                U(x_i) = \\left\\{ \\frac{1}{n-1} \\sum_{j \\in [n] \\setminus \\{i\\}} U_1(x_i , x_j)\\right\\}


            The ranking for the given query set :math:`Q` is defined as:

            .. math::

                ρ(Q)  = \\operatorname{argsort}_{x \\in Q}  \\; U(x)


           Parameters
           ----------
           n_hidden : int
               Number of hidden layers used in the scoring network
           n_units : int
               Number of hidden units in each layer of the scoring network
           loss_function : function or string
               Loss function to be used for the binary decision task of the
               pairwise comparisons
           batch_normalization : bool
               Whether to use batch normalization in each hidden layer
           kernel_regularizer : function
               Regularizer function applied to all the hidden weight matrices.
           activation : function or string
               Type of activation function to use in each hidden layer
           optimizer : function or string
               Optimizer to use during stochastic gradient descent
           metrics : list
               List of metrics to evaluate during training (can be
               non-differentiable)
           batch_size : int
               Batch size to use during training
           random_state : int, RandomState instance or None
               Seed of the pseudorandom generator or a RandomState instance
           **kwargs
               Keyword arguments for the algorithms


           References
           ----------
                [1] Leonardo Rigutini, Tiziano Papini, Marco Maggini, and Franco Scarselli. 2011. SortNet: Learning to Rank by a Neural Preference Function. IEEE Trans. Neural Networks 22, 9 (2011), 1368–1380. https://doi.org/10.1109/TNN.2011.2160875
        """
        super().__init__(
            n_hidden=n_hidden,
            n_units=n_units,
            loss_function=loss_function,
            batch_normalization=batch_normalization,
            kernel_regularizer=kernel_regularizer,
            kernel_initializer=kernel_initializer,
            activation=activation,
            optimizer=optimizer,
            metrics=metrics,
            batch_size=batch_size,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(CmpNet.__name__)
        self.logger.info("Initializing network")

    def _convert_instances_(self, X, Y):
        self.logger.debug("Creating the Dataset")
        garbage, x1, x2, y_double, garbage = generate_complete_pairwise_dataset(X, Y)
        del garbage
        if x1.shape[0] > self.threshold_instances:
            indices = self.random_state_.choice(
                x1.shape[0], self.threshold_instances, replace=False
            )
            x1 = x1[indices, :]
            x2 = x2[indices, :]
            y_double = y_double[indices, :]
        self.logger.debug("Finished the Dataset instances {}".format(x1.shape[0]))
        return x1, x2, y_double

    def construct_model(self):
        return super().construct_model()

    def fit(
        self, X, Y, epochs=10, callbacks=None, validation_split=0.1, verbose=0, **kwd
    ):
        """
            Fit an object ranking learning CmpNet model on a provided set of queries. The provided queries can be of a
            fixed size (numpy arrays). For learning this network the binary cross entropy loss function for a pair of
            objects :math:`x_i, x_j \\in Q` is defined as:

            .. math::

                C_{ij} =  -\\tilde{P_{ij}}(0)\\cdot \\log(U(x_i,x_j)) - \\tilde{P_{ij}}(1) \\cdot \\log(U(x_j,x_i)) \\ ,

            where :math:`\\tilde{P_{ij}}` is ground truth probability of the preference of :math:`x_i` over :math:`x_j`.
            :math:`\\tilde{P_{ij}} = (1,0)` if :math:`x_i \\succ x_j` else :math:`\\tilde{P_{ij}} = (0,1)`.

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Rankings of the given objects
            epochs : int
                Number of epochs to run if training for a fixed query size
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float
                Percentage of instances to split off to validate on
            verbose : bool
                Print verbose information
            **kwd
                Keyword arguments for the fit function
        """
        super().fit(
            X,
            Y,
            epochs=epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            verbose=verbose,
            **kwd,
        )

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ObjectRanker.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def clear_memory(self, **kwargs):
        super().clear_memory(**kwargs)

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden=n_hidden,
            n_units=n_units,
            reg_strength=reg_strength,
            learning_rate=learning_rate,
            batch_size=batch_size,
            **point,
        )



__version__ = "1.2.0"

# We should re-evaluate if we really want to re-export everything here and then
# use __all__ properly.

from .choicefunction import *  # noqa
from .core import *  # noqa
from .dataset_reader import *  # noqa
from .discretechoice import *  # noqa
from .objectranking import *  # noqa
from .tunable import Tunable  # noqa
from .tuning import ParameterOptimizer  # noqa



from collections import namedtuple
import os
import re

import numpy as np
from scipy.stats import rankdata
from sklearn.utils import check_random_state

from csrank.constants import OBJECT_RANKING
from .util import sub_sampling_rankings
from ..dataset_reader import DatasetReader

__all__ = ["DepthDatasetReader"]


class DepthDatasetReader(DatasetReader):
    def __init__(self, dataset_type="deep", random_state=None, **kwargs):
        super(DepthDatasetReader, self).__init__(
            learning_problem=OBJECT_RANKING, dataset_folder="depth_data", **kwargs
        )
        options = {
            "deep": ["complete_deep_train.dat", "complete_deep_test.dat"],
            "basic": ["saxena_basic61x55.dat", "saxena_basicTest61x55.dat"],
            "semantic": ["saxena_semantic61x55.dat", "saxena_semanticTest61x55.dat"],
        }
        if dataset_type not in options:
            raise ValueError(f"dataset_type must be one of set({options.keys()})")
        train_filename, test_file_name = options[dataset_type]
        self.train_file = os.path.join(self.dirname, train_filename)
        self.test_file = os.path.join(self.dirname, test_file_name)
        self.random_state = check_random_state(random_state)

        self.__load_dataset__()

    def __load_dataset__(self):
        self.x_train, self.depth_train = load_dataset(self.train_file)
        self.x_test, self.depth_test = load_dataset(self.test_file)

    def get_train_test_datasets(self, n_datasets=5):
        splits = np.array(n_datasets)
        return self.splitter(splits)

    def get_single_train_test_split(self):
        seed = self.random_state.randint(2 ** 32, dtype="uint32")
        X_train, Y_train = self.get_train_dataset_sampled_partial_rankings(seed=seed)
        X_train, Y_train = sub_sampling_rankings(X_train, Y_train, n_objects=5)
        X_test, Y_test = self.get_test_dataset_ties()
        return X_train, Y_train, X_test, Y_test

    def get_dataset_dictionaries(self):
        pass

    def splitter(self, iter):
        for i in iter:
            X_train, Y_train = self.get_train_dataset_sampled_partial_rankings(
                seed=10 * i + 32
            )
            X_test, Y_test = self.get_test_dataset_ties()
        yield X_train, Y_train, X_test, Y_test

    def get_test_dataset_sampled_partial_rankings(self, **kwargs):
        self.X, self.Y = self.get_dataset_sampled_partial_rankings(
            datatype="test", **kwargs
        )
        self.__check_dataset_validity__()
        return self.X, self.Y

    def get_train_dataset_sampled_partial_rankings(self, **kwargs):
        self.X, self.Y = self.get_dataset_sampled_partial_rankings(
            datatype="train", **kwargs
        )
        self.__check_dataset_validity__()
        return self.X, self.Y

    def get_test_dataset(self):
        self.X, self.Y = self.get_dataset(datatype="test")
        self.__check_dataset_validity__()
        return self.X, self.Y

    def get_train_dataset(self):
        self.X, self.Y = self.get_dataset(datatype="train")
        self.__check_dataset_validity__()
        return self.X, self.Y

    def get_test_dataset_ties(self):
        self.X, self.Y = self.get_dataset_ties(datatype="test")
        self.__check_dataset_validity__()
        return self.X, self.Y

    def get_train_dataset_ties(self):
        self.X, self.Y = self.get_dataset_ties(datatype="train")
        self.__check_dataset_validity__()
        return self.X, self.Y

    def get_dataset_sampled_partial_rankings(
        self, datatype="train", max_number_of_rankings_per_image=10, seed=42
    ):
        random_state = np.random.RandomState(seed=seed)
        x_train, depth_train = self.get_deep_copy_dataset(datatype)
        X = []
        rankings = []
        order_lengths = np.array(
            [
                len(
                    np.unique(depths[np.where(depths <= 0.80)[0]], return_index=True)[1]
                )
                for depths in depth_train
            ]
        )
        order_length = np.min(order_lengths)

        for features, depths in zip(x_train, depth_train):
            value, obj_indices = np.unique(
                depths[np.where(depths <= 0.80)[0]], return_index=True
            )
            interval = int(len(obj_indices) / order_length)
            if interval < max_number_of_rankings_per_image:
                num_of_orderings_per_image = interval
            else:
                num_of_orderings_per_image = max_number_of_rankings_per_image
            objects_i = np.empty([order_length, num_of_orderings_per_image], dtype=int)
            for i in range(order_length):
                if i != order_length - 1:
                    objs = random_state.choice(
                        obj_indices[i * interval : (i + 1) * interval],
                        num_of_orderings_per_image,
                        replace=False,
                    )
                else:
                    objs = random_state.choice(
                        obj_indices[i * interval : len(obj_indices)],
                        num_of_orderings_per_image,
                        replace=False,
                    )
                objects_i[i] = objs
            for i in range(num_of_orderings_per_image):
                indices = objects_i[:, i]
                np.random.shuffle(indices)
                X.append(features[indices])
                ranking = rankdata(depths[indices]) - 1
                rankings.append(ranking)
        X = np.array(X)
        rankings = np.array(rankings)
        return X, rankings

    def get_dataset(self, datatype="train"):
        x, y = self.get_deep_copy_dataset(datatype)
        X = []
        rankings = []
        for features, depths in zip(x, y):
            value, indices = np.unique(depths, return_index=True)
            np.random.shuffle(indices)
            X.append(features[indices])
            ranking = rankdata(depths[indices]) - 1
            rankings.append(ranking)
        X = np.array(X)
        rankings = np.array(rankings)
        return X, rankings

    def get_dataset_ties(self, datatype="train"):
        X, y = self.get_deep_copy_dataset(datatype)
        for depth in y:
            depth[np.where(depth >= 0.80)[0]] = 0.80
        rankings = np.array([rankdata(depth) - 1 for depth in y])
        return X, rankings

    def get_deep_copy_dataset(self, datatype):
        if datatype == "train":
            x, y = np.copy(self.x_train), np.copy(self.depth_train)
        elif datatype == "test":
            x, y = np.copy(self.x_test), np.copy(self.depth_test)
        return x, y


def load_dataset(filename):
    Instance = namedtuple("Instance", ["depth", "features"])
    instances = dict()
    with open(filename) as f:
        for line in f:
            arr = line.split()
            depth = float(arr[0])
            qid = int(arr[1][4:])
            if qid not in instances:
                instances[qid] = []
                if "#" in arr:
                    arr = arr[:-2]
            features = [float(re.search(r"\:([0-9\.]*)", x).group(1)) for x in arr[2:]]
            instances[qid].append(Instance(depth, features))
    n_instances = len(instances)
    n_objects = len(instances[1])
    n_features = len(instances[1][0][1])
    X = np.empty((n_instances, n_objects, n_features))
    y = np.empty((n_instances, n_objects))
    for i, inst in enumerate(instances.values()):
        for j, (depth, features) in enumerate(inst):
            X[i, j] = features
            y[i, j] = depth
        ind = np.argsort(y[i, :])
        y[i] = y[i, ind]
        X[i] = X[i, ind]
    return X, y



from abc import ABCMeta
from abc import abstractmethod

from csrank.constants import DYAD_RANKING

__all__ = ["DyadRanker"]


class DyadRanker(metaclass=ABCMeta):
    @property
    def learning_problem(self):
        return DYAD_RANKING

    @abstractmethod
    def fit(self, Xo, Xc, Y, **kwargs):
        """ Fit the contextual object ranking algorithm for object features
         Xo and context features Xc and orderings Y for the objects.

        Parameters
        ----------
        Xo : array-like, shape (n_samples, n_objects, n_object_features)
            Feature vectors of the objects
        Xc : array-like, shape (n_samples, n_context_features)
            Feature vector of the context
        Y : array-like, shape (n_samples, n_objects)
            Orderings of the objects

        Returns
        -------
        self : object
            Returns self.
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, Xo, Xc, **kwargs):
        """ Predict orderings for the objects for a given context vector.

        Parameters
        ----------
        Xo : array-like, shape (n_samples, n_objects, n_object_features)
        Xc : array-like, shape (n_samples, n_context_features)


        Returns
        -------
        Y : array-like, shape (n_samples, n_objects)
            Predicted orderings
        """
        raise NotImplementedError

    @abstractmethod
    def predict_scores(self, Xo, Xc, **kwargs):
        """ Predict scores for a given collection of sets of objects.

        Parameters
        ----------
        Xo : array-like, shape (n_samples, n_objects, n_object_features)
        Xc : array-like, shape (n_samples, n_context_features)


        Returns
        -------
        Y : array-like, shape (n_samples, n_objects)
            Returns the scores of each of the objects for each of the
            samples.
        """
        raise NotImplementedError

    def __call__(self, Xo, Xc, *args, **kwargs):
        """ Predicts orderings for a new set of objects Xo.

        Parameters
        ----------
        Xo : array-like, shape (n_samples, n_objects, n_object_features)
        Xc : array-like, shape (n_samples, n_context_features)

        Returns
        -------
        Y : array-like, shape (n_samples, n_objects)
            Predicted orderings"""
        return self.predict(Xo, Xc, **kwargs)

    @classmethod
    def __subclasshook__(cls, C):
        if cls is DyadRanker:
            has_fit = any("fit" in B.__dict__ for B in C.__mro__)
            has_predict = any("predict" in B.__dict__ for B in C.__mro__)
            has_scores = any("predict_scores" in B.__dict__ for B in C.__mro__)
            if has_fit and has_predict and has_scores:
                return True
        return NotImplemented



import logging

from keras.losses import binary_crossentropy
from sklearn.model_selection import train_test_split

from csrank.core.fate_linear import FATELinearCore
from .choice_functions import ChoiceFunctions


class FATELinearChoiceFunction(FATELinearCore, ChoiceFunctions):
    def __init__(
        self,
        n_hidden_set_units=2,
        loss_function=binary_crossentropy,
        learning_rate=1e-3,
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """
            Create a FATELinear-network architecture for leaning discrete choice function. The first-aggregate-then-evaluate
            approach learns an embedding of each object and then aggregates that into a context representation
            :math:`\\mu_{C(x)}` and then scores each object :math:`x` using a generalized utility function
            :math:`U (x, \\mu_{C(x)})`.
            To make it computationally efficient we take the the context :math:`C(x)` as query set :math:`Q`.
            The context-representation is evaluated as:

            .. math::
                \\mu_{C(x)} = \\frac{1}{\\lvert C(x) \\lvert} \\sum_{y \\in C(x)} \\phi(y)

            where :math:`\\phi \\colon \\mathcal{X} \\to \\mathcal{Z}` maps each object :math:`y` to an
            :math:`m`-dimensional embedding space :math:`\\mathcal{Z} \\subseteq \\mathbb{R}^m`.
            Training complexity is quadratic in the number of objects and prediction complexity is only linear.
            The discrete choice for the given query set :math:`Q` is defined as:

            .. math::

                dc(Q) := \\operatorname{argmax}_{x \\in Q}  \\;  U (x, \\mu_{C(x)})

            Parameters
            ----------
            n_hidden_set_units : int
                Number of hidden set units.
            batch_size : int
                Batch size to use for training
            loss_function : function
                Differentiable loss function for the score vector
            random_state : int or object
                Numpy random state
            **kwargs
                Keyword arguments for the @FATENetwork
        """
        super().__init__(
            n_hidden_set_units=n_hidden_set_units,
            learning_rate=learning_rate,
            batch_size=batch_size,
            loss_function=loss_function,
            random_state=random_state,
            **kwargs,
        )
        self.logger = logging.getLogger(FATELinearChoiceFunction.__name__)

    def fit(
        self,
        X,
        Y,
        epochs=10,
        callbacks=None,
        validation_split=0.1,
        tune_size=0.1,
        thin_thresholds=1,
        verbose=0,
        **kwd,
    ):
        if tune_size > 0:
            X_train, X_val, Y_train, Y_val = train_test_split(
                X, Y, test_size=tune_size, random_state=self.random_state
            )
            try:
                super().fit(
                    X_train,
                    Y_train,
                    epochs,
                    callbacks,
                    validation_split,
                    verbose,
                    **kwd,
                )
            finally:
                self.logger.info(
                    "Fitting utility function finished. Start tuning threshold."
                )
                self.threshold = self._tune_threshold(
                    X_val, Y_val, thin_thresholds=thin_thresholds, verbose=verbose
                )
        else:
            super().fit(X, Y, epochs, callbacks, validation_split, verbose, **kwd)
            self.threshold = 0.5

    def _predict_scores_fixed(self, X, **kwargs):
        return super()._predict_scores_fixed(X, **kwargs)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ChoiceFunctions.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def set_tunable_parameters(
        self,
        n_hidden_set_units=32,
        learning_rate=1e-3,
        batch_size=128,
        epochs_drop=300,
        drop=0.1,
        **point,
    ):
        super().set_tunable_parameters(
            n_hidden_set_units=n_hidden_set_units,
            learning_rate=learning_rate,
            batch_size=batch_size,
            epochs_drop=epochs_drop,
            drop=drop,
            **point,
        )



import os

from keras.optimizers import SGD
import numpy as np
from pymc3.variational.callbacks import CheckParametersConvergence
import pytest
import tensorflow as tf

from csrank.choicefunction import *
from csrank.constants import CMPNET_CHOICE
from csrank.constants import FATE_CHOICE
from csrank.constants import FATELINEAR_CHOICE
from csrank.constants import FETA_CHOICE
from csrank.constants import FETALINEAR_CHOICE
from csrank.constants import GLM_CHOICE
from csrank.constants import RANKNET_CHOICE
from csrank.constants import RANKSVM_CHOICE
from csrank.metrics_np import auc_score
from csrank.metrics_np import f1_measure
from csrank.metrics_np import instance_informedness
from csrank.metrics_np import subset_01_loss
from csrank.tests.test_ranking import check_params_tunable
from csrank.util import metrics_on_predictions

choice_metrics = {
    "F1Score": f1_measure,
    "Informedness": instance_informedness,
    "AucScore": auc_score,
}
optimizer = SGD(lr=1e-3, momentum=0.9, nesterov=True)


def get_vals(values):
    return dict(zip(choice_metrics.keys(), values))


choice_functions = {
    FETA_CHOICE: (
        FETAChoiceFunction,
        {"add_zeroth_order_model": True, "optimizer": optimizer},
        get_vals([0.946, 0.9684, 0.9998]),
    ),
    FATE_CHOICE: (
        FATEChoiceFunction,
        {
            "n_hidden_joint_layers": 1,
            "n_hidden_set_layers": 1,
            "n_hidden_joint_units": 5,
            "n_hidden_set_units": 5,
            "optimizer": optimizer,
        },
        get_vals([0.8185, 0.6070, 0.9924]),
    ),
    FATELINEAR_CHOICE: (
        FATELinearChoiceFunction,
        {},
        get_vals([0.8014, 0.4906, 0.9998]),
    ),
    FETALINEAR_CHOICE: (
        FETALinearChoiceFunction,
        {},
        get_vals([0.8782, 0.8894, 0.9998]),
    ),
    RANKNET_CHOICE: (
        RankNetChoiceFunction,
        {"optimizer": optimizer},
        get_vals([0.9522, 0.9866, 1.0]),
    ),
    CMPNET_CHOICE: (
        CmpNetChoiceFunction,
        {"optimizer": optimizer},
        get_vals([0.8554, 0.8649, 0.966]),
    ),
    GLM_CHOICE: (GeneralizedLinearModel, {}, get_vals([0.9567, 0.9955, 1.0])),
    RANKSVM_CHOICE: (PairwiseSVMChoiceFunction, {}, get_vals([0.9522, 0.9955, 1.0])),
}


@pytest.fixture(scope="module")
def trivial_choice_problem():
    random_state = np.random.RandomState(42)
    x = random_state.randn(200, 5, 1)
    y_true = np.array(x.squeeze(axis=-1) > np.mean(x))
    return x, y_true


@pytest.mark.parametrize("name", list(choice_functions.keys()))
def test_choice_function_fixed(trivial_choice_problem, name):
    tf.set_random_seed(0)
    os.environ["KERAS_BACKEND"] = "tensorflow"
    np.random.seed(123)
    x, y = trivial_choice_problem
    choice_function = choice_functions[name][0]
    params, accuracies = choice_functions[name][1], choice_functions[name][2]
    learner = choice_function(**params)
    if name == GLM_CHOICE:
        learner.fit(
            x,
            y,
            vi_params={
                "n": 100,
                "method": "advi",
                "callbacks": [CheckParametersConvergence()],
            },
        )
    elif "linear" in name:
        learner.fit(x, y, epochs=10, validation_split=0, verbose=False)
    else:
        learner.fit(x, y, epochs=100, validation_split=0, verbose=False)
    s_pred = learner.predict_scores(x)
    y_pred = learner.predict_for_scores(s_pred)
    y_pred_2 = learner.predict(x)
    rtol = 1e-2
    atol = 5e-2
    assert np.isclose(
        0.0, subset_01_loss(y_pred, y_pred_2), rtol=rtol, atol=atol, equal_nan=False
    )
    for key, value in accuracies.items():
        metric = choice_metrics[key]
        if metric in metrics_on_predictions:
            pred_loss = metric(y, y_pred)
        else:
            pred_loss = metric(y, s_pred)
        assert np.isclose(value, pred_loss, rtol=rtol, atol=atol, equal_nan=False)
    params = {
        "n_hidden": 20,
        "n_units": 20,
        "n_hidden_set_units": 2,
        "n_hidden_set_layers": 10,
        "n_hidden_joint_units": 2,
        "n_hidden_joint_layers": 10,
        "reg_strength": 1e-3,
        "learning_rate": 1e-1,
        "batch_size": 32,
        "alpha": 0.5,
        "l1_ratio": 0.7,
        "tol": 1e-2,
        "C": 10,
        "n_mixtures": 10,
        "n_nests": 5,
        "regularization": "l2",
    }
    learner.set_tunable_parameters(**params)
    check_params_tunable(learner, params, rtol, atol)



import logging
import multiprocessing
import os

from keras import backend as K
import numpy as np
import tensorflow as tf
from tensorflow.python.client import device_lib


def scores_to_rankings(n_objects, y_pred):
    # indices = orderings
    toprel, orderings = tf.nn.top_k(y_pred, n_objects)
    # indices = rankings
    troprel, rankings = tf.nn.top_k(orderings, n_objects)
    rankings = K.cast(rankings[:, ::-1], dtype="float32")
    return rankings


def get_instances_objects(y_true):
    n_objects = K.cast(K.int_shape(y_true)[1], "int32")
    total = K.cast(K.greater_equal(y_true, 0), dtype="int32")
    n_instances = K.cast(tf.reduce_sum(total) / n_objects, dtype="int32")
    return n_instances, n_objects


def tensorify(x):
    """Converts x into a Keras tensor"""
    if not isinstance(x, (tf.Tensor, tf.Variable)):
        return K.constant(x)
    return x


def get_tensor_value(x):
    if isinstance(x, tf.Tensor):
        return K.get_value(x)
    return x


def configure_numpy_keras(seed=42):
    tf.set_random_seed(seed)
    os.environ["KERAS_BACKEND"] = "tensorflow"
    devices = [x.name for x in device_lib.list_local_devices()]
    logger = logging.getLogger("ConfigureKeras")
    logger.info("Devices {}".format(devices))
    n_gpus = len(
        [x.name for x in device_lib.list_local_devices() if x.device_type == "GPU"]
    )
    if n_gpus == 0:
        config = tf.ConfigProto(
            intra_op_parallelism_threads=1,
            inter_op_parallelism_threads=1,
            allow_soft_placement=True,
            log_device_placement=False,
            device_count={"CPU": multiprocessing.cpu_count() - 2},
        )
    else:
        config = tf.ConfigProto(
            allow_soft_placement=True,
            log_device_placement=True,
            intra_op_parallelism_threads=2,
            inter_op_parallelism_threads=2,
        )  # , gpu_options = gpu_options)
    sess = tf.Session(config=config)
    K.set_session(sess)
    np.random.seed(seed)
    logger.info("Number of GPUS {}".format(n_gpus))


def get_mean_loss(metric, y_true, y_pred):
    if isinstance(y_pred, dict) and isinstance(y_true, dict):
        losses = []
        total_instances = 0
        for n in y_pred.keys():
            loss = eval_loss(metric, y_true[n], y_pred[n])
            if not np.isnan(loss) and not np.isinf(loss):
                loss = loss * y_pred[n].shape[0]
                total_instances += y_pred[n].shape[0]
                losses.append(loss)
        losses = np.array(losses)
        mean_loss = np.sum(losses) / total_instances
    else:
        mean_loss = eval_loss(metric, y_true, y_pred)
    return mean_loss


def eval_loss(metric, y_true, y_pred):
    x = metric(y_true, y_pred)
    x = get_tensor_value(x)
    return np.nanmean(x)



import logging

from keras import backend as K
from keras import Input
from keras import optimizers
from keras.layers import concatenate
from keras.layers import Dense
from keras.models import Model
from keras.optimizers import SGD
from keras.regularizers import l2
from sklearn.utils import check_random_state
import tensorflow as tf

from csrank.constants import allowed_dense_kwargs
from csrank.layers import create_input_lambda
from csrank.layers import NormalizedDense
from csrank.learner import Learner
from csrank.losses import plackett_luce_loss
from csrank.metrics import zero_one_rank_loss_for_scores_ties
from csrank.objectranking.object_ranker import ObjectRanker
from csrank.util import print_dictionary

__all__ = ["ListNet"]


class ListNet(Learner, ObjectRanker):
    def __init__(
        self,
        n_top=1,
        n_hidden=2,
        n_units=8,
        loss_function=plackett_luce_loss,
        batch_normalization=False,
        kernel_regularizer=l2(1e-4),
        activation="selu",
        kernel_initializer="lecun_normal",
        optimizer=SGD(lr=1e-4, nesterov=True, momentum=0.9),
        metrics=[zero_one_rank_loss_for_scores_ties],
        batch_size=256,
        random_state=None,
        **kwargs,
    ):
        """ Create an instance of the ListNet architecture. ListNet trains a latent utility model based on
            top-k-subrankings of the objects. This network learns a latent utility score for each object in the given
            query set :math:`Q = \\{x_1, \\ldots ,x_n\\}` using the equation :math:`U(x) = F(x, w)` where :math:`w` is the
            weight vector. A listwise loss function like the negative Plackett-Luce likelihood is used for training.
            The ranking for the given query set :math:`Q` is defined as:

            .. math::

                ρ(Q)  = \\operatorname{argsort}_{x \\in Q}  \\; U(x)

            Parameters
            ----------
            n_top : int
                Size of the top-k-subrankings to consider for training
            hash_file: str
                File path of the model where the weights are stored to get the predictions after clearing the memory
            n_hidden : int
                Number of hidden layers used in the scoring network
            n_units : int
                Number of hidden units in each layer of the scoring network
            loss_function : function or string
                Listwise loss function which is applied on the top-k objects
            batch_normalization : bool
                Whether to use batch normalization in each hidden layer
            kernel_regularizer : function
                Regularizer function applied to all the hidden weight matrices.
            activation : function or string
                Type of activation function to use in each hidden layer
            kernel_initializer : function or string
                Initialization function for the weights of each hidden layer
            optimizer : function or string
                Optimizer to use during stochastic gradient descent
            metrics : list
                List of metrics to evaluate during training (can be
                non-differentiable)
            batch_size : int
                Batch size to use during training
            random_state : int, RandomState instance or None
                Seed of the pseudorandom generator or a RandomState instance
            **kwargs
                Keyword arguments for the algorithms

            References
            ----------
                [1] Z. Cao, T. Qin, T. Liu, M. Tsai and H. Li. "Learning to Rank: From Pairwise Approach to Listwise Approach." ICML, 2007.
        """
        self.logger = logging.getLogger(ListNet.__name__)
        self.n_top = n_top
        self.batch_normalization = batch_normalization
        self.activation = activation
        self.metrics = metrics
        self.kernel_regularizer = kernel_regularizer
        self.kernel_initializer = kernel_initializer
        self.loss_function = loss_function
        self.optimizer = optimizers.get(optimizer)
        self._optimizer_config = self.optimizer.get_config()
        self.n_hidden = n_hidden
        self.n_units = n_units
        keys = list(kwargs.keys())
        for key in keys:
            if key not in allowed_dense_kwargs:
                del kwargs[key]
        self.kwargs = kwargs

        self.threshold_instances = int(1e10)
        self.batch_size = batch_size
        self.random_state = random_state
        self.hash_file = None
        self.model = None
        self._scoring_model = None

    def _construct_layers(self, **kwargs):
        self.input_layer = Input(shape=(self.n_top, self.n_object_features_fit_))
        self.output_node = Dense(
            1, activation="linear", kernel_regularizer=self.kernel_regularizer
        )
        if self.batch_normalization:
            self.hidden_layers = [
                NormalizedDense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        else:
            self.hidden_layers = [
                Dense(self.n_units, name="hidden_{}".format(x), **kwargs)
                for x in range(self.n_hidden)
            ]
        assert len(self.hidden_layers) == self.n_hidden

    def _create_topk(self, X, Y):
        n_inst, n_obj, n_feat = X.shape
        mask = Y < self.n_top
        X_topk = X[mask].reshape(n_inst, self.n_top, n_feat)
        Y_topk = Y[mask].reshape(n_inst, self.n_top)
        return X_topk, Y_topk

    def fit(
        self, X, Y, epochs=10, callbacks=None, validation_split=0.1, verbose=0, **kwd
    ):
        """
            Fit an object ranking learning ListNet on the top-k-subrankings in the provided set of queries. The provided
            queries can be of a fixed size (numpy arrays). For fitting the model we maximize the Plackett-Luce
            likelihood. For example for query set :math:`Q = \\{x_1,x_2,x_3\\}`, the scores are :math:`Q = (s_1,s_2,s_3)`
            and the ranking is :math:`\\pi = (3,1,2)`. The  Plackett-Luce likelihood is defined as:

            .. math::
                P_l(\\pi) = \\frac{s_2}{s_1+s_2+s_3} \\cdot \\frac{s_3}{s_1+s_3} \\cdot \\frac{s_1}{s_1}

            Note: For k=2 we obtain :class:`RankNet` as a special case.

            Parameters
            ----------
            X : numpy array
                (n_instances, n_objects, n_features)
                Feature vectors of the objects
            Y : numpy array
                (n_instances, n_objects)
                Rankings of the given objects
            epochs : int
                Number of epochs to run if training for a fixed query size
            callbacks : list
                List of callbacks to be called during optimization
            validation_split : float
                Percentage of instances to split off to validate on
            verbose : bool
                Print verbose information
            **kwd
                Keyword arguments for the fit function
        """
        self.random_state_ = check_random_state(self.random_state)
        _n_instances, _n_objects, self.n_object_features_fit_ = X.shape
        self._construct_layers(
            kernel_regularizer=self.kernel_regularizer,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation,
            **self.kwargs,
        )
        self.logger.debug("Creating top-k dataset")
        X, Y = self._create_topk(X, Y)
        self.logger.debug("Finished creating the dataset")

        self.logger.debug("Creating the model")
        self.model = self.construct_model()
        self.logger.debug("Finished creating the model, now fitting...")
        self.model.fit(
            X,
            Y,
            batch_size=self.batch_size,
            epochs=epochs,
            callbacks=callbacks,
            validation_split=validation_split,
            verbose=verbose,
            **kwd,
        )
        self.logger.debug("Fitting Complete")

    def construct_model(self):
        """
            Construct the ListNet architecture which takes topk-subrankings from the given queries and minimize a
            listwise loss on the utility scores of top objects. Weight sharing guarantees that we learn the shared
            weights :math:`w` of the latent utility function :math:`U(x) = F(x, w)`.

            Returns
            -------
            model: keras model :class:`Model`
                ListNet model used to learn the utiliy function using the top-k-subrankings in the provided set of queries.
        """
        hid = [create_input_lambda(i)(self.input_layer) for i in range(self.n_top)]
        for hidden_layer in self.hidden_layers:
            hid = [hidden_layer(x) for x in hid]
        outputs = [self.output_node(x) for x in hid]
        merged = concatenate(outputs)
        model = Model(inputs=self.input_layer, outputs=merged)
        model.compile(
            loss=self.loss_function, optimizer=self.optimizer, metrics=self.metrics
        )
        return model

    @property
    def scoring_model(self):
        """
            Creates a scoring model from the trained ListNet, which predicts the utility scores for given set of objects.
            This network consist of a sequential network which predicts the utility score for each object :math:`x \\in Q`
            using the latent utility function :math:`U(x) = F(x, w)` where :math:`w` is the weights of the model.

            Returns
            -------
            scoring_model: keras model :class:`Model`
                scoring model used to predict utility score for each object
        """
        if self._scoring_model is None:
            self.logger.info("Creating scoring model")
            inp = Input(shape=(self.n_object_features_fit_,))
            x = inp
            for hidden_layer in self.hidden_layers:
                x = hidden_layer(x)
            output_score = self.output_node(x)
            self._scoring_model = Model(inputs=inp, outputs=output_score)
        return self._scoring_model

    def _predict_scores_fixed(self, X, **kwargs):
        n_inst, n_obj, n_feat = X.shape
        self.logger.info(
            "For Test instances {} objects {} features {}".format(*X.shape)
        )
        inp = Input(shape=(n_obj, n_feat))
        lambdas = [create_input_lambda(i)(inp) for i in range(n_obj)]
        scores = concatenate([self.scoring_model(lam) for lam in lambdas])
        model = Model(inputs=inp, outputs=scores)
        return model.predict(X)

    def predict_scores(self, X, **kwargs):
        return super().predict_scores(X, **kwargs)

    def predict_for_scores(self, scores, **kwargs):
        return ObjectRanker.predict_for_scores(self, scores, **kwargs)

    def predict(self, X, **kwargs):
        return super().predict(X, **kwargs)

    def clear_memory(self, **kwargs):
        """
            Clear the memory, restores the currently fitted model back to prevent memory leaks.

            Parameters
            ----------
            **kwargs :
                Keyword arguments for the function
        """
        if self.hash_file is not None:
            self.model.save_weights(self.hash_file)
            K.clear_session()
            sess = tf.Session()
            K.set_session(sess)
            self._scoring_model = None
            self.optimizer = self.optimizer.from_config(self._optimizer_config)
            self._construct_layers(
                kernel_regularizer=self.kernel_regularizer,
                kernel_initializer=self.kernel_initializer,
                activation=self.activation,
                **self.kwargs,
            )
            self.model = self.construct_model()
            self.model.load_weights(self.hash_file)
        else:
            self.logger.info("Cannot clear the memory")

    def set_tunable_parameters(
        self,
        n_hidden=32,
        n_units=2,
        reg_strength=1e-4,
        learning_rate=1e-3,
        batch_size=128,
        **point,
    ):
        """
            Set tunable parameters of the ListNet network to the values provided.

            Parameters
            ----------
            n_hidden: int
                Number of hidden layers used in the scoring network
            n_units: int
                Number of hidden units in each layer of the scoring network
            reg_strength: float
                Regularization strength of the regularizer function applied to the `kernel` weights matrix
            learning_rate: float
                Learning rate of the stochastic gradient descent algorithm used by the network
            batch_size: int
                Batch size to use during training
            point: dict
                Dictionary containing parameter values which are not tuned for the network
        """
        self.n_hidden = n_hidden
        self.n_units = n_units
        self.kernel_regularizer = l2(reg_strength)
        self.batch_size = batch_size
        self.optimizer = self.optimizer.from_config(self._optimizer_config)
        K.set_value(self.optimizer.lr, learning_rate)
        self._construct_layers(
            kernel_regularizer=self.kernel_regularizer,
            kernel_initializer=self.kernel_initializer,
            activation=self.activation,
            **self.kwargs,
        )

        self._scoring_model = None
        if len(point) > 0:
            self.logger.warning(
                "This ranking algorithm does not support "
                "tunable parameters called: {}".format(print_dictionary(point))
            )
